# -*- coding: utf-8 -*-

# Copyright © 2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import numpy as np
from tramway.tessellation.base import *
import scipy.spatial as spatial
from collections import namedtuple, deque, defaultdict


_Voronoi = namedtuple('BoxedVoronoi', (
        'points',
        'vertices',
        'ridge_points',
        'ridge_vertices',
        'regions',
        'point_region',
    ))

def boxed_voronoi_2d(points, bounding_box=None):
    """ moved from the *base* module with version *0.5.2*; *deprecated* """
    voronoi = spatial.Voronoi(points)
    if bounding_box is None:
        #x_min = np.minimum(np.min(points, axis=0), np.min(voronoi.vertices, axis=0))
        #x_max = np.maximum(np.max(points, axis=0), np.max(voronoi.vertices, axis=0))
        x_min, x_max = np.min(points, axis=0), np.max(points, axis=0)
        dx = x_max - x_min
        x_min -= 1e-4 * dx
        x_max += 1e-4 * dx
        bounding_box = (
            x_min,
            np.array([x_max[0], x_min[1]]),
            x_max,
            np.array([x_min[0], x_max[1]]),
            )
    t = (bounding_box[0] + bounding_box[2]) * .5
    lstsq_kwargs = dict(rcond=None)
    region_point = np.full(len(voronoi.regions), -1, dtype=int)
    region_point[voronoi.point_region] = np.arange(voronoi.point_region.size)
    extra_vertices = []
    vertex_index = n_vertices = voronoi.vertices.shape[0]
    extra_ridges = []
    n_ridges = voronoi.ridge_points.shape[0]
    _ridge_vertices, _ridge_points, _regions = \
        voronoi.ridge_vertices, voronoi.ridge_points, [[]]
    for c, u in enumerate(points):
        r = voronoi.point_region[c]
        region = voronoi.regions[r]
        #assert region_point[r] == c
        _region = []
        for k, h in enumerate(region):
            if 0 <= h:
                _region.append(h)
                continue

            # find the two "adjacent" vertices
            i, j = region[k-1], region[(k+1)%len(region)]
            assert 0<=i
            assert 0<=j
            # find the corresponding neighbour cells
            m, n = set(), set() # mutable
            for d, hs in enumerate(voronoi.regions):
                if not hs or d == r:
                    continue
                for e, cs in ((i,m), (j,n)):
                    try:
                        l = hs.index(e)
                    except ValueError:
                        continue
                    if (hs[l-1]==-1) or (hs[(l+1)%len(hs)]==-1):
                        cs.add(region_point[d])
            p, q = m.pop(), n.pop()
            assert not m
            assert not n
            # pick a distant point on the perpendicular bissector
            # of the neighbour centers, at the intersection with
            # the bounding box
            prev_edge = None
            for h, d in ((i,p), (j,q)):
                gs = []
                try:
                    gs = _ridge_vertices[n_ridges+extra_ridges.index([d,c])]
                except ValueError:
                    try:
                        gs = _ridge_vertices[n_ridges+extra_ridges.index([c,d])]
                    except ValueError:
                        pass
                if gs:
                    g, = [ g for g in gs if g != h ]
                    _region.append(g)
                    continue
                v = voronoi.vertices[h]
                w = points[d]
                if np.any(v<bounding_box[0]) or np.any(bounding_box[2]<v):
                    # vertex v stands outside the bounding box
                    # TODO: check for corners as potential intermediate vertices
                    continue
                n = np.array([u[1]-w[1], w[0]-u[0]])
                #y = (u + w) * .5
                y = v
                # determine direction: n or -n?
                if 0 < np.dot(n, t-y):
                    n = -n
                # intersection of [y,n) and [a,ab]
                v_next = None
                for l, a in enumerate(bounding_box):
                    b = bounding_box[(l+1)%len(bounding_box)]
                    M, p = np.c_[n, a-b], a-y
                    q = np.linalg.lstsq(M, p, **lstsq_kwargs)
                    q = q[0]
                    if 0<=q[0] and 0<=q[1] and q[1]<=1:
                        # intersection found
                        v_next = y + q[0] * n
                        #if 0<q[0]:
                        break
                assert v_next is not None
                if prev_edge is None or prev_edge == l:
                    h_prev = h
                else:
                    # add the corner which the previous
                    # and current edges intersect at
                    e_max = len(bounding_box)-1
                    if (0<l and prev_edge==l-1) or \
                        (l==0 and prev_edge==e_max):
                        v_prev = a
                    elif (l<e_max and prev_edge==l+1) or \
                        (l==e_max and prev_edge==0):
                        v_prev = b
                    else:
                        # two corners?
                        # better connect the intersection points and let the corners out
                        h_prev = h
                        v_prev = None
                    if v_prev is not None:
                        h_prev = vertex_index
                        vertex_index += 1
                        extra_vertices.append(v_prev)
                        extra_ridges.append([-1,c])
                        _ridge_vertices.append([h,h_prev])
                        _region.append(h_prev)
                prev_edge = l
                #
                h_next = vertex_index
                vertex_index += 1
                # insert the new vertex
                extra_vertices.append(v_next)
                k, = ((_ridge_points[:,0]==c) & (_ridge_points[:,1]==d)).nonzero()
                if k.size==0:
                    k, = ((_ridge_points[:,0]==d) & (_ridge_points[:,1]==c)).nonzero()
                k = k[0].tolist()
                _ridge_vertices = _ridge_vertices[:k] \
                    + [[h_prev,h_next]] \
                    + _ridge_vertices[k+1:]
                _region.append(h_next)
                assert len(extra_vertices) == vertex_index - n_vertices

        assert all( r in _region for r in region if 0 < r )
        _regions.append(_region)
    _points = voronoi.points
    _vertices = np.vstack([voronoi.vertices]+extra_vertices)
    _ridge_points = np.vstack([_ridge_points]+extra_ridges)
    _point_region = np.arange(1, len(_regions))#voronoi.point_region
    return _Voronoi(_points, _vertices, _ridge_points, _ridge_vertices, _regions, _point_region)


def get_exterior_cells(tessellation, relative_margin=.05, bounds=None, extended_cell_filter=None):
    """
    Looks for Voronoi cells that expand outside the 2D data bounding box.

    This function can be called in two ways:

    * pass a partition object, which bounding box is inferred from location data,
      and optionaly define a margin on this default bounding box,
    * pass a SciPy *Voronoi* or tessellation object and explicitly define the bounding box.

    Arguments:

        tessellation (*scipy.spatial.Voronoi* or *Voronoi* or *Partition*):
            SciPy *Voronoi* object,
            or spatial segmentation (*TimeLattice* objects are supported),
            or full data partition

        relative_margin (float):
            margin for the default bounds, to identify outside vertices

        bounds (pair of *numpy.ndarray*):
            lower and upper bounds on the spatial coordinates

        extended_cell_filter (callable):
            boolean function that takes the vertices of any closed cell with at least
            one out-of-bound vertex, and returns :const:`True` if the corresponding
            cell should be marked as exterior, :const:`False` otherwise

    Returns:

        list: indices of exterior cells
    
    Modified from https://github.com/DecBayComp/Stochastic_Integrals_Diffusivity/blob/master/ito-to-tramway/get_exterior_cells.py
    """
    if bounds is None:
        if isinstance(tessellation, Partition):
            partition = tessellation
            bb = partition.bounding_box
            space_cols = list('xy')
            lb, ub = bb.loc['min',space_cols].values, bb.loc['max',space_cols].values
        else:
            raise ValueError('undefined bounds')
    else:
        lb, ub = bounds

    if isinstance(tessellation, spatial.Voronoi):
        voronoi = tessellation

    else:
        if isinstance(tessellation, Partition):
            tessellation = tessellation.tessellation
        try:
            tessellation = tessellation.spatial_mesh
        except AttributeError:
            pass
        # rebuild the Voronoi diagram with explicit undefined vertices (with index -1)
        voronoi = spatial.Voronoi(tessellation.cell_centers)

    margin = relative_margin * (ub-lb)
    outside_vertices = np.any((tessellation.vertices<lb-margin)|(ub+margin<tessellation.vertices), axis=1)

    exterior_cells = set()
    ncells = len(voronoi.points)

    for cell_ix in range(ncells):

        region_ix = voronoi.point_region[cell_ix]
        if region_ix < 0:
            continue

        vertex_ids = np.asarray(voronoi.regions[region_ix])

        if np.any(vertex_ids < 0):
            # divergent cell
            exterior_cells.add(cell_ix)

        elif np.any(outside_vertices[vertex_ids]):
            # out-of-bounds vertex
            if not extended_cell_filter or extended_cell_filter(voronoi.vertices[vertex_ids]):
                exterior_cells.add(cell_ix)

    return list(exterior_cells)


class Path(object):
    __slots__ = ('fragments',)
    def __init__(self):
        self.fragments = []
    def add_edge(self, i, j):
        """
        If edge *(i,j)* closes the path, `add_edge` returns the resulting path.
        Otherwise :const:`None` is returned instead.
        """
        connected = None
        if self.fragments:
            fragments = []
            connected_twice = False
            for f, fragment in enumerate(self.fragments):
                if connected_twice:
                    fragments.append(fragment)
                    continue
                #
                ends, path = fragment
                # check if the (i, j) edge connects with the `fragment`
                if i in ends:
                    resume_from, continuation = i, j
                elif j in ends:
                    resume_from, continuation = j, i
                else:
                    # the edge does not connect;
                    # leave the fragment untouched
                    fragments.append(fragment)
                    continue
                # sanity checks
                if continuation in path:
                    if continuation in ends and not self.fragments[1:]:
                        # the path is complete
                        _, complete_path = self.fragments[0]
                        return complete_path
                    else:
                        raise ValueError('duplicate edges')
                elif connected is None:
                    # connect the (i,j) edge with the path `fragment`:
                    # replace the `resume_from` node by the `continuation` node
                    ends = ends - {resume_from} | {continuation}
                    # append the `continuation` node to the `path`
                    if path[0] == resume_from:
                        path.appendleft(continuation)
                    else: #if path[-1] == resume_from
                        path.append(continuation)
                    # update the fragment
                    fragments.append((ends, path))
                    connected = f
                else:
                    # combine fragments
                    fragments, _fragments = [], fragments
                    for f, fragment in enumerate(_fragments):
                        if f == connected:
                            ends0, path0 = fragment
                            connection_point = resume_from
                            # check `ends` and `ends0` connect
                            if not (connection_point in ends and connection_point in ends0):
                                raise RuntimeError('triangle found')
                            assert continuation in path0
                            # connect `path` with `path0`
                            new_ends = ends0 ^ ends
                            assert len(new_ends)==2
                            if path[0] != connection_point: # or path[-1] == connection_point
                                path.reverse()
                            path.popleft()
                            new_path = path0
                            if path0[0] == connection_point:
                                new_path.extendleft(path)
                            else:# if path0[-1] == connection_point
                                new_path.extend(path)
                            fragment = (new_ends, new_path)
                        fragments.append(fragment)
                    connected_twice = True
            self.fragments = fragments
        if connected is None:
            # initialize a new path fragment
            self.fragments.append((set((i,j)), deque((i,j))))
    def get_path(self):
        if len(self.fragments) == 1:
            _, path = self.fragments[0]
            path = list(path)
        else:
            path = None
        return path


def get_interior_contour(tessellation, relative_margin=None, bounds=None, return_indices=False,
        extended_cell_filter=None):
    """
    Calls `get_exterior_cells` and returns the 2D inner vertices of exterior cells
    so that these vertices make a contour around the interior cells.

    Can return vertex indices instead.
    Input arguments are similar to those of `get_interior_contour`.
    """
    if bounds is None and isinstance(tessellation, Partition):
        partition = tessellation
        bb = partition.bounding_box
        space_cols = list('xy')
        bounds = bb.loc['min',space_cols].values, bb.loc['max',space_cols].values

    if isinstance(tessellation, spatial.Voronoi):
        voronoi = tessellation

    else:
        if isinstance(tessellation, Partition):
            tessellation = tessellation.tessellation
        try:
            tessellation = tessellation.spatial_mesh
        except AttributeError:
            pass
        # rebuild the Voronoi diagram with explicit undefined vertices (with index -1)
        voronoi = spatial.Voronoi(tessellation.cell_centers)

    if relative_margin is None:
        outer_ids = get_exterior_cells(voronoi, bounds=bounds,
                extended_cell_filter=extended_cell_filter)
    else:
        outer_ids = get_exterior_cells(voronoi, relative_margin,
                bounds, extended_cell_filter)

    outer_ids = set(outer_ids)

    adjacency = defaultdict(set)
    for k, ij in enumerate(voronoi.ridge_points):
        i, j = ij
        adjacency[i].add((j, k))
        adjacency[j].add((i, k))

    contour_ridge_ids = set()
    for i in outer_ids:
        contour_ridge_ids |= { k for j, k in adjacency[i] if j not in outer_ids }

    vertex_path = Path()
    while contour_ridge_ids:
        ridge_ix = contour_ridge_ids.pop()
        if vertex_path.add_edge(*voronoi.ridge_vertices[ridge_ix]):
            assert not contour_ridge_ids
    vertex_path = vertex_path.get_path()

    if return_indices:
        return vertex_path
    else:
        return voronoi.vertices[vertex_path]


__all__ = ['_Voronoi', 'boxed_voronoi_2d', \
    'get_exterior_cells', 'get_interior_contour']




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import scipy.sparse as sparse
from .exception import *


class Graph(object):
    '''Abstract class for node-centered graph implementations. Here the concept of edge is limited to
    labels and retrieving the nodes that a given edge connects may be a very inefficient operation.'''
    __slots__ = ['node_defaults', 'edge_defaults']

    def __init__(self, node_defaults, edge_defaults):
        self.node_defaults = node_defaults
        self.edge_defaults = edge_defaults

    def connect(self, n1, n2, **kwargs):
        raise AbstractGraphError

    def disconnect(self, n1, n2, edge=None):
        raise AbstractGraphError

    def get_node_attr(self, n, attr):
        raise AbstractGraphError
    def set_node_attr(self, n, **kwargs):
        raise AbstractGraphError
    def get_edge_attr(self, e, attr):
        raise AbstractGraphError
    def set_edge_attr(self, e, **kwargs):
        raise AbstractGraphError
    @property
    def size(self):
        return len(list(self.iter_nodes))
    def iter_nodes(self):
        '''Returns an iterator over all the nodes of the graph. A copy of the returned
        view should be made if the structure of the graph had to be altered.'''
        raise AbstractGraphError
    def iter_neighbors(self, n):
        '''Returns an iterator over the neighbor nodes of a given node. A copy of the returned
        view should be made if the structure of the graph had to be altered.'''
        return [ n for _, n in self.iter_edges_from(n) ]
    def iter_edges(self):
        '''Returns an iterator over all the edges of the graph. A copy of the returned
        view should be made if the structure of the graph had to be altered.'''
        raise AbstractGraphError
    def iter_edges_from(self, n):
        '''Returns an iterator over the ``(edge, neighbor_node)`` index pairs from a given node.
        A copy of the returned view should be made if the structure of the graph had to be altered.'''
        raise AbstractGraphError
    def has_node(self, n):
        return n in self.iter_nodes()

    def add_node(self, **kwargs):
        raise AbstractGraphError
    def del_node(self, n):
        raise AbstractGraphError
    def stands_alone(self, n):
        '''Returns ``True`` if node `n` has no neighbor.'''
        return not self.iter_neighbors(n)

    def find_edge(self, n1, n2):
        if self.has_node(n1):
            e = [ e for e, n in self.iter_edges_from(n1) if n is n2 ]
            if e: return e[0]
            else: return None
        else:
            return None

    def are_connected(self, n1, n2):
        return self.find_edge(n1, n2) is not None

    def export(self, sparse='csr'):
        """Returns ``(A, V, E)`` where ``A`` is an adjacency matrix as a `scipy.sparse.*_matrix`
        of size ``(n_vertices, n_vertices)`` and with edge indices as defined values (including
        explicit zeros), where * is defined by the optional input argument `sparse` (any of
        '*csc*', '*csr*', '*bsr*', '*lil*', '*dok*', '*coo*', '*dia*').
        ``V`` is a `dict` with ``n_vertex_attributes`` entries and where each value is a
        ``(n_vertices, *)`` matrix (usually vector).
        ``E`` is a `dict` with ``n_edge_attributes`` entries and where each value is a
        ``(n_edges, *)`` matrix (usually vector).
        Note that vertices and edges may be differently ordered compared to the current `Graph`
        implementation. Especially, ``V`` and ``E`` should be dense."""
        raise AbstractGraphError

    def squareDistance(self, attr, eta, **kwargs):
        raise AbstractGraphError




# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os.path
from .base import *
from .gradient import *
from tramway.core.plugin import Plugins


plugins = Plugins(os.path.dirname(__file__), __package__,
        {'infer': 'infer.*'}, require='setup')

# add `worker_count` argument to every mode with `cell_sampling` option set
# and `dilation` and `max_cell_count` if `cell_sampling` is not 'individual'
_wc_args = dict(type=int, help='number of parallel processes to spawn')
_mcc_args = dict(type=int, default=999999999999999999, help='max number of cells per group')
_dil_args = dict(type=int, help='cell group overlap as a number of "layers" of cells')
_sigma_args = dict(type=float, help='localization precision (distance)')
_sigma2_args = dict(type=float, help='localization error (distance square)')
def _post_load(plugins):
    for _mode in plugins:
        _setup, _module = plugins[_mode]
        _args = _setup.get('arguments', {})
        if 'cell_sampling' in _setup:
            if _args:
                if 'worker_count' in _args:
                    continue
                _flags = [ _a[0] for _a in _args.values()
                        if isinstance(_a, (tuple, list)) and _a ]
            else:
                _setup['arguments'] = {}
                _flags = []
            if '-w' in _flags:
                _setup['arguments']['worker_count'] = _wc_args
            else:
                _setup['arguments']['worker_count'] = ('-w', _wc_args)
            if _setup['cell_sampling'] != 'individual':
                if '-C' in _flags:
                    _setup['arguments']['max_cell_count'] = _mcc_args
                else:
                    _setup['arguments']['max_cell_count'] = ('-C', _mcc_args)
                _setup['arguments']['dilation'] = _dil_args
        if 'localization_error' in _args:
            if 'sigma' not in _args:
                _setup['arguments']['sigma'] = _sigma_args
            if 'sigma2' not in _args:
                _setup['arguments']['sigma2'] = _sigma2_args
        if 'diffusivity_prior' in _args:
            _dprior = _args['diffusivity_prior']
            assert isinstance(_dprior, tuple)
            assert _dprior[0] == '-d'
            _dprior = (_dprior[0], '--diffusion-prior') + _dprior[1:]
            _setup['arguments']['diffusivity_prior'] = _dprior
        if 'diffusivity_time_prior' in _args:
            _dprior = _args['diffusivity_time_prior']
            assert isinstance(_dprior, tuple)
            assert isinstance(_dprior[0], str)
            _dprior = (_dprior[0], '--diffusion-time-prior') + _dprior[1:]
            _setup['arguments']['diffusivity_time_prior'] = _dprior

plugins.post_load = _post_load




# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Temporary file where stuff usefull for supervised classification is put,
to try and save datasets.
"""

import multiprocessing as mp

import numpy as np
import pandas as pd
import torch
import torch.utils.data
import tqdm
from torch import nn, optim
from torch.nn import functional as F

from .rw_features import *
from .batch_generation import create_batch_rw, generate_random_number


def interpolate(X1, X2, alpha):
    return alpha * X1 + (1-alpha) * X2


def get_interpolation(t, x, tnew, old=False):
    N = len(t)
    Nnew = len(tnew)
    xnew = np.zeros((Nnew, x.shape[1]))
    dt = t[1] - t[0]
    dtnew = tnew[1] - tnew[0]
    for i in range(Nnew):
        if old:
            j = min(int(dtnew / dt * i), N-2)
            alpha = (t[j+1] - tnew[i]) / dt
            xnew[i] = interpolate(x[j], x[j+1], alpha)
        else:
            j = min(int(dtnew / dt * i), N-1)
            xnew[i] = x[j]
    return xnew


def normalize_length(RW, N, old=False):
    """Performs a regularization of the number of points of a random walk.
    As long as the random walk has at least 2 points, this function is able to
    transform it to a random walk with an arbitrary large number of points.
    Two mode are possible :
       - old = True : interpolates between points
       - old = False : keeps the last point if super sampling.
    """
    tnew = np.linspace(0, RW.t.max(), num=N, endpoint=True)
    Xnew = get_interpolation(RW.t.values, RW.loc[:, ['x', 'y']].values,
                             tnew, old=old)
    data = np.stack((tnew,) + tuple(Xnew[:, i] for i in range(2))).T
    return pd.DataFrame(data=data, columns=['t'] + ['x', 'y'])


def process_rw(args):
    """This function is used to create the raw data used in the supervised
    classification task. It is made to be compatible with multiprocessing of
    different random walks.

    Parameters
    ----------
    args : tuple of rw : random walk DataFrame
    id_ : dummy id variable of the random walk
    kwargs : dictionary of meta parameters. Contains a list of windows
        ('windows' key) and the maximum size of this list ('wmax' key).

    Returns
    -------
    RW_ws : a dict with keys the different windows used to resample the random
        walk. For each window, we first normalize the length of the random
        walk. Then for x and y, we divide by the mean step and the total
        distance covered to get 4 channels. We add two more channels : the
        cumulative distance covered rescaled between 0 and 1 and the mean
        square displacement function.
    scale_prm : a dict {window_size : dt(window_size)}.
    meta_prm : parameters of the random walk containing its scaling : Tmax,
        mean step size, and distance covered.
    id_ : dummy id of the random walk.
    """
    rw, id_, kwargs = args
    length_rw = len(rw)
    dt = rw.t.iloc[1] - rw.t.iloc[0]
    X = rw.loc[:, ['x', 'y']].values
    t = rw.t.values
    rw_obj = RandomWalk(rw)
    steps = np.diagonal(rw_obj.Dabs, offset=1)
    mean_step = np.mean(steps)
    sum_step = np.sum(steps)
    cum_dist = np.insert(
        np.cumsum(steps / sum_step), 0, 0)
    tau, msd = temporal_msd(
        rw_obj, n_samples=kwargs['wmax'], use_all=False)
    msd /= (np.max(msd) if np.max(msd) > 0 else 1)
    n_msd = len(msd)
    RW_ws = {}
    scale_prm = {}
    for k, w in enumerate(kwargs['windows']):
        RW_w = normalize_length(rw, w)
        RW_w.loc[:, ['x', 'y']] /= mean_step
        RW_w['x_perc'] = RW_w.loc[:, 'x'] / sum_step
        RW_w['y_perc'] = RW_w.loc[:, 'y'] / sum_step
        interp_f = scipy.interpolate.interp1d(
            t, cum_dist, fill_value=(min(cum_dist), max(cum_dist)))
        RW_w['cum_dist'] = interp_f(RW_w.t.values)
        interp_f = scipy.interpolate.interp1d(
            tau, msd, fill_value=(min(msd), max(msd)))
        RW_w['msd_log'] = interp_f(np.linspace(min(tau), max(tau), num=w))
        RW_ws[w] = RW_w.loc[:, [
            'x', 'y', 'x_perc', 'y_perc', 'cum_dist', 'msd_log']].values.T
        scale_prm[w] = {'dt': dt * length_rw / w}
    meta_prm = {'Tmax': rw.t.max(), 'mean_step': mean_step,
                'sum_step': sum_step}
    return RW_ws, scale_prm, meta_prm, id_


class get_label_from_func():
    """Class that provides a call method to get an 'y' out of parameters of the
    random walk.
    """

    def __init__(self, types):
        self.types_name = list(set([x.__name__ for x, _ in types]))
        nt = len(self.types_name)
        self.labels = []
        for i in range(len(self.types_name)):
            v = np.zeros(nt)
            v[i] = 1
            self.labels.append(v)
        self.dict_val_label = dict(list(zip(self.types_name, self.labels)))

    def __call__(self, prm):
        return self.dict_val_label[prm['func_name']]


class RWDataset(torch.utils.data.Dataset):
    """torch Dataset subclass.
    """

    def __init__(self, RWs, prms, windows=[100, 10], nb_process=4,
                 y_func='from_func', **kwargs):
        self.RWs = RWs
        self.RWsgroup = RWs.groupby('n')
        self.prms = prms
        self.windows = windows
        self.w_max = max(windows)
        self.trajs = RWs.n.unique()
        self.dict_index_traj = dict(
            list(zip(np.arange(len(self.trajs)), self.trajs)))
        self.RW_ws, self.scale_prm = dict(
            (w, {}) for w in windows), dict((w, {}) for w in windows)
        self.meta_prm = {}
        if y_func == 'from_func':
            self.get_y_func = get_label_from_func(kwargs['types'])

        rws = [(rw, id_, {'windows': self.windows, 'wmax': self.w_max})
               for id_, rw in self.RWs.groupby('n')]
        desc = 'distributed work'
        n = len(self.trajs)
        if nb_process is None:
            raw_data = list(map(process_rw,
                                tqdm.tqdm_notebook(rws, total=n, desc=desc)))
        else:
            with mp.Pool(nb_process) as p:
                raw_data = list(tqdm.tqdm_notebook(p.imap(process_rw, rws),
                                                   total=n, desc=desc))

        for i in tqdm.tqdm_notebook(range(len(raw_data)), desc='concat'):
            RW_ws_i, scale_prm_i, meta_prm_i, id_ = raw_data[i]
            self.meta_prm[id_] = meta_prm_i
            for k, w in enumerate(self.windows):
                self.RW_ws[w][id_] = RW_ws_i[w]
                self.scale_prm[w][id_] = scale_prm_i[w]

    def __len__(self):
        return len(self.trajs)

    def __getitem__(self, index):
        traj = self.dict_index_traj[index]
        rws = tuple([torch.tensor(self.RW_ws[w][traj]).float()
                     for w in self.windows])
        scale_prms = tuple([torch.tensor(
            list(self.scale_prm[w][traj].values())).float()
            for w in self.windows])
        meta_prms = torch.tensor(list(self.meta_prm[traj].values())).float()
        y = self.get_y_func(self.prms[traj])
        return rws, scale_prms, meta_prms, torch.tensor(y).float()


class conv_layer(nn.Module):
    """Performs 1d convolutions on a 1D Tensor and outputs a 1D activation
    Tensor using global average pooling.
    """

    def __init__(self, w, scale_size=1, out_size=50, input_channel=1):
        super(conv_layer, self).__init__()
        self.w = w
        wtemp = self.w
        k_size = 7 if (wtemp > 70) else (
            5 if (wtemp <= 70 and wtemp > 30) else 3)
        temp_conv = []
        channels_in = input_channel
        channels_out = 16
        while wtemp > k_size:
            k_size = 7 if (wtemp > 50) else (
                5 if (wtemp < 50 and wtemp > 15) else 3)
            temp_conv.extend([nn.Conv1d(in_channels=channels_in,
                                        out_channels=channels_out,
                                        kernel_size=k_size),
                              nn.ReLU(), nn.MaxPool1d(2)])
            wtemp = (wtemp - 2 * k_size // 2) / 2
            channels_in = channels_out
            channels_out *= 2
        temp_conv.append(nn.AdaptiveAvgPool1d(1))
        self.conv = nn.Sequential(*temp_conv)
        self.fc_scaler = nn.Linear(channels_in + scale_size, out_size)

    def forward(self, RW, scale_prm):
        out_rw = torch.squeeze(self.conv(RW), dim=2)
        out_rw.view(out_rw.size(0), -1)
        return F.relu(self.fc_scaler(torch.cat((out_rw, scale_prm), dim=1)))


class RWNet(nn.Module):
    """Main neural network. It performs 1D convolutions separately on each
    channel of the provided temporal data. It adds scaling factors and finally
    meta variables before last layers.
    """

    def __init__(self, hidden_dims, supervised=True, windows=[100, 10],
                 scale_size=1, out_cnn_size=40, meta_size=2, input_channels=6,
                 output_size_softmax=1, output_size_linear=0):
        super(RWNet, self).__init__()
        self.hidden_dims = hidden_dims
        self.output_size_softmax = output_size_softmax
        self.output_size_linear = output_size_linear
        self.output_size = self.output_size_softmax + self.output_size_linear
        self.supervised = supervised
        self.windows = windows
        self.input_channels = input_channels

        self.conv_layers = {}
        for w in self.windows:
            for channel in range(self.input_channels):
                self.conv_layers[(w, channel)] = conv_layer(
                    w, scale_size=scale_size, out_size=out_cnn_size)

        enc_layers = [nn.Linear((out_cnn_size * len(self.windows) *
                                 self.input_channels) + meta_size,
                                self.hidden_dims[0]),
                      nn.ReLU()]
        lin_enc = [nn.Linear(self.hidden_dims[i], self.hidden_dims[i+1])
                   for i in range(len(self.hidden_dims)-1)]
        for i in range(len(lin_enc)):
            enc_layers += [lin_enc[i], nn.ReLU()]
        self.fc = nn.Sequential(*enc_layers)
        self.fc_mu = nn.Linear(self.hidden_dims[-1], self.output_size)
        if not supervised:
            self.fc_logvar = nn.Linear(self.hidden_dims[-1], self.output_size)

    def forward(self, rws, scale_prms, meta_prms):
        feat_in_fc = torch.cat((tuple(
            self.conv_layers[(w, channel)](rws[i][:, [channel], :],
                                           scale_prms[i])
            for i, w in enumerate(self.windows)
            for channel in range(self.input_channels)) +
            (meta_prms,)), dim=1)
        feat_out_fc = self.fc(feat_in_fc)
        if self.supervised:
            if self.output_size_linear == 0:
                return torch.squeeze(F.softmax(self.fc_mu(feat_out_fc), dim=1))
            else:
                raw_activ = self.fc_mu(feat_out_fc)
                return torch.cat((
                    F.softmax(raw_activ[:, :self.output_size_softmax]),
                    raw_activ[:, self.output_size_softmax:]), dim=1)
        else:
            return self.fc_mu(feats), self.fc_logvar(feats)


def process_rw2(args):
    rw, id_, kwargs = args
    length_rw = len(rw)
    dt = rw.t.iloc[1] - rw.t.iloc[0]
    X = rw.loc[:, ['x', 'y']].values
    t = rw.t.values
    rw_obj = RandomWalk(rw)
    steps = np.diagonal(rw_obj.Dabs, offset=1)
    mean_step = np.mean(steps)
    sum_step = np.sum(steps)
    cum_dist = np.insert(
        np.cumsum(steps / sum_step), 0, 0)
    tau, msd = temporal_msd(
        rw_obj, use_all=False, n_samples=kwargs['n_samples'])
    msd /= (np.max(msd) if np.max(msd) > 0 else 1)
    n_msd = len(msd)
    RW_ws = {}
    try:
        interp_cum = scipy.interpolate.interp1d(
            t, cum_dist, fill_value=(min(cum_dist), max(cum_dist)))
        cum_dist = interp_cum(np.linspace(
            0, t.max(), num=kwargs['n_samples'], endpoint=True))
    except:
        interp_cum = scipy.interpolate.interp1d(
            t, cum_dist, fill_value='extrapolate')
        cum_dist = interp_cum(np.linspace(
            0, t.max(), num=kwargs['n_samples'], endpoint=True))
    try:
        interp_msd = scipy.interpolate.interp1d(
            tau, msd, fill_value=(min(msd), max(msd)))
        msd = interp_msd(np.geomspace(
            min(tau), max(tau), num=kwargs['n_samples']))
    except:
        interp_msd = scipy.interpolate.interp1d(
            tau, msd, fill_value='extrapolate')
        msd = interp_msd(np.geomspace(
            min(tau), max(tau), num=kwargs['n_samples']))
    for k, w in enumerate(kwargs['windows']):
        X = normalize_length(rw, w).loc[:, ['x', 'y']].values
        X /= mean_step
        RW_ws[w] = X.T
    meta_prm = np.array([t.max() - t.min(), mean_step, sum_step])
    return RW_ws, id_, msd, cum_dist, meta_prm


class get_y_binary():
    def __init__(self, func1):
        self.func1 = func1

    def __call__(self, prm):
        return 1 if prm['func_name'] == self.func1 else 0


class RWDatasetTmp3(torch.utils.data.Dataset):
    """torch Dataset subclass.
    """

    def __init__(self, RWs, prms, kwargs, nb_process=16,
                 func_feat_y=get_y_binary('RW_FBM')):
        self.ns = RWs.n.unique()
        self.RWsgroup = RWs.groupby('n')
        self.prms = prms

        self.kwargs = kwargs
        self.func_feat_y = func_feat_y

        rws = [(rw, id_, kwargs)
               for id_, rw in self.RWsgroup]

        if nb_process is None:
            raw_data = list(map(process_rw2,
                                tqdm.tqdm_notebook(rws, total=len(self.ns))))
        else:
            with mp.Pool(nb_process) as p:
                raw_data = list(tqdm.tqdm_notebook(
                    p.imap(process_rw2, rws), total=len(self.ns)))
        self.RWsdict = {}
        self.msds = {}
        self.cum_dists = {}
        self.meta_prms = {}
        self.dict_index_n = {}
        for i in tqdm.tqdm_notebook(range(len(raw_data))):
            RW_ws, id_, msd, cum_dist, meta_prm = raw_data[i]
            self.dict_index_n[i] = id_
            self.RWsdict[id_] = {w: torch.tensor(
                RW_ws[w]).float() for w in RW_ws.keys()}
            self.msds[id_] = torch.tensor(msd).float()
            self.cum_dists[id_] = torch.tensor(cum_dist).float()
            self.meta_prms[id_] = torch.tensor(meta_prm).float()

    def __len__(self):
        return len(self.dict_index_n)

    def __getitem__(self, index):
        traj = self.dict_index_n[index]
        y = torch.tensor(self.func_feat_y(self.prms[traj])).float()
        return (self.RWsdict[traj], self.msds[traj], self.cum_dists[traj],
                self.meta_prms[traj], y)


def get_cum_hull(RW, nb_samples):
    positions = np.unique(RW.position, axis=0)
    if len(positions) > 3:
        hull = scipy.spatial.ConvexHull(positions[:3], incremental=True)
        cum_hull = [hull.volume]
        ts = np.unique(np.geomspace(3, len(positions), num=nb_samples,
                                    dtype=int, endpoint=True))
    #     print(ts)
        i0 = 3
        for j, i in enumerate(ts[1:]):
            hull.add_points(positions[i0:i])
            cum_hull.append(hull.volume)
            i0 = i
        cum_hull = np.array(cum_hull) / cum_hull[-1]
        interp_cv = scipy.interpolate.interp1d(ts, cum_hull,
                                               fill_value='extrapolate')
        new_ts = np.geomspace(3, len(positions), num=nb_samples, endpoint=True)
    #     print(new_ts)
        cv_hull_regu = interp_cv(new_ts)
        return new_ts, cv_hull_regu
    else:
        return (np.geomspace(3, 300, num=nb_samples, endpoint=True),
                np.geomspace(0.0001, 1, num=nb_samples, endpoint=True))


def process_rw3(args):
    rw, id_, kwargs = args
    length_rw = len(rw)
    dt = rw.t.iloc[1] - rw.t.iloc[0]
    X = rw.loc[:, ['x', 'y']].values
    if kwargs['pos_noise'] is not None:
        noise = generate_random_number(*kwargs['pos_noise'])
        X += np.random.randn(*X.shape) * noise
    t = rw.t.values
    rw_obj = RandomWalk(rw)
    steps = np.diagonal(rw_obj.Dabs, offset=1)
    mean_step = np.mean(steps)
    sum_step = np.sum(steps)
    cum_dist = np.insert(
        np.cumsum(steps / sum_step), 0, 0)
    tau, msd = temporal_msd(
        rw_obj, use_all=False, n_samples=kwargs['n_samples'])
    msd /= (np.max(msd) if np.max(msd) > 0 else 1)
    n_msd = len(msd)
    RW_ws = {}
    interp_cum = scipy.interpolate.interp1d(
        t, cum_dist, fill_value='extrapolate')
    cum_dist = interp_cum(np.linspace(
        0, t.max(), num=kwargs['n_samples'], endpoint=True))
    interp_msd = scipy.interpolate.interp1d(
        tau, msd, fill_value='extrapolate')
    msd = interp_msd(np.geomspace(
        min(tau), max(tau), num=kwargs['n_samples']))
    cv_hull = get_cum_hull(rw_obj, kwargs['n_samples'])[1]
    for k, w in enumerate(kwargs['windows']):
        X = normalize_length(rw, w).loc[:, ['x', 'y']].values
        X /= mean_step
        RW_ws[w] = X.T
    meta_prm = np.array(list(rw_obj.get_all_features().values()))
    return RW_ws, id_, msd, cum_dist, cv_hull, meta_prm


class RWDatasetTmp4(torch.utils.data.Dataset):
    """torch Dataset subclass.
    """

    def __init__(self, RWs, prms, kwargs, nb_process=16, chunksize=40,
                 func_feat_y=get_y_binary('RW_FBM')):
        self.ns = RWs.n.unique()
        self.RWsgroup = RWs.groupby('n')
        self.prms = prms

        self.kwargs = kwargs
        self.func_feat_y = func_feat_y

        rws = [(rw, id_, kwargs)
               for id_, rw in self.RWsgroup]

        if nb_process is None:
            raw_data = list(map(process_rw3,
                                tqdm.tqdm_notebook(rws, total=len(self.ns))))
        else:
            with mp.Pool(nb_process) as p:
                raw_data = list(tqdm.tqdm_notebook(
                    p.imap(process_rw3, rws, chunksize), total=len(self.ns)))
        self.RWsdict = {}
        self.msds = {}
        self.cum_dists = {}
        self.cv_hulls = {}
        self.meta_prms = {}
        self.dict_index_n = {}
        meta_prms = np.array([raw_data[i][5] for i in range(len(raw_data))])
        self.mean_meta = np.mean(meta_prms, axis=0)
        self.std_meta = np.std(meta_prms, axis=0)
        self.std_meta[self.std_meta < 1e-6] = 1
        for i in tqdm.tqdm_notebook(range(len(raw_data))):
            RW_ws, id_, msd, cum_dist, cv_hull, meta_prm = raw_data[i]
            self.dict_index_n[i] = id_
            self.RWsdict[id_] = {w: torch.tensor(
                RW_ws[w]).float() for w in RW_ws.keys()}
            self.msds[id_] = torch.tensor(msd).float()
            self.cum_dists[id_] = torch.tensor(cum_dist).float()
            self.cv_hulls[id_] = torch.tensor(cv_hull).float()
            self.meta_prms[id_] = torch.tensor((meta_prm - self.mean_meta) /
                                               self.std_meta).float()

    def __len__(self):
        return len(self.dict_index_n)

    def __getitem__(self, index):
        traj = self.dict_index_n[index]
        y = torch.tensor(self.func_feat_y(self.prms[traj])).float()
        return (self.RWsdict[traj], self.msds[traj], self.cum_dists[traj],
                self.cv_hulls[traj], self.meta_prms[traj], y)


def make_datasets(PATH, types, ps=None, M=50, N=10000, nb_process=16,
                  funcDataset=RWDatasetTmp3, kwargs={}, nb_pos_min=3,
                  chunk_size=100):
    output_paths = []
    ds_lengths = []
    for i in tqdm.tqdm_notebook(range(M)):
        RWs, prms = create_batch_rw(n=N, nb_process=nb_process, types=types,
                                    ps=ps, nb_pos_min=nb_pos_min,
                                    chunksize=chunk_size)
        ds = funcDataset(RWs, prms, kwargs, nb_process=nb_process,
                         chunksize=chunk_size)
        ds_lengths.append(len(ds))
        torch.save(ds, f'{PATH}_file_{i}')
        output_paths.append(f'{PATH}_file_{i}')
    return output_paths, ds_lengths



# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent
# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import *
import numpy as np
import numpy.linalg as la
from .graph import *
#from .graph.array import ArrayGraph
import time
from scipy.spatial.distance import cdist
from scipy.special import gamma
from .dichotomy import Dichotomy


class Gas(Graph):
    """Implementation of the *Grow(ing) When Required* clustering algorithm, first inspired from
    [Marsland02]_ and then extensively modified.

    .. [Marsland02] Marsland, S., Shapiro, J. and Nehmzow, U. (2002). A self-organising network that grows when required. Neural Networks 15, 1041-1058. doi:10.1026/S0893-6080(02)00078-3

    Of note, :class:`Gas` features a :attr:`trust` attribute that enables faster learning when
    inserting new nodes. Instead of inserting a node at ``(w + eta) / 2`` (with ``trust == 0``) like
    in the standard algorithm, you can insert sample point `eta` (with ``trust == 1``) or any point
    in between.
    :class:`Gas` also features a 'collapse' step at the end of each batch. This step consists of
    merging nodes that are closer from each other than threshold :attr:`collapse_below`.
    Last but not least, the insertion threshold and collapse threshold can be functions of the
    current training sample and nearest node.

    Implementation note: :class:`Gas` implements :class:`Graph` as a proxy, instanciating or taking
    another `Graph` object with optional input argument `graph` and delegates all :class:`Graph` 's
    method to that object attribute.
    """
    __slots__ = ['graph', 'insertion_threshold', 'trust', 'learning_rate', \
        'habituation_threshold', 'habituation_initial', 'habituation_alpha', \
        'habituation_tau', 'edge_lifetime', 'batch_size', 'collapse_below', 'knn', \
        'topology']

    def connect(self, n1, n2, **kwargs):
        self.graph.connect(n1, n2, **kwargs)
    def disconnect(self, n1, n2, edge=None):
        self.graph.disconnect(n1, n2, edge)
    def get_node_attr(self, n, attr):
        return self.graph.get_node_attr(n, attr)
    def set_node_attr(self, n, **kwargs):
        self.graph.set_node_attr(n, **kwargs)
    def get_edge_attr(self, e, attr):
        return self.graph.get_edge_attr(e, attr)
    def set_edge_attr(self, e, **kwargs):
        self.graph.set_edge_attr(e, **kwargs)
    @property
    def size(self):
        return self.graph.size
    def iter_nodes(self):
        return self.graph.iter_nodes()
    def iter_neighbors(self, n):
        return self.graph.iter_neighbors(n)
    def iter_edges(self):
        return self.graph.iter_edges()
    def iter_edges_from(self, n):
        return self.graph.iter_edges_from(n)
    def has_node(self, n):
        return self.graph.has_node(n)
    def add_node(self, **kwargs):
        return self.graph.add_node(**kwargs)
    def del_node(self, n):
        self.graph.del_node(n)
    def stands_alone(self, n):
        return self.graph.stands_alone(n)
    def find_edge(self, n1, n2):
        return self.graph.find_edge(n1, n2)
    def are_connected(self, n1, n2):
        return self.graph.are_connected(n1, n2)
    def export(self, **kwargs):
        return self.graph.export(**kwargs)
    def square_distance(self, attr, eta, **kwargs):
        return self.graph.square_distance(attr, eta, **kwargs)

    def __init__(self, sample, graph=None):
        if 1 < sample.shape[0]:
            w1 = sample[0]
            w2 = sample[-1]
        else:
            raise ValueError
        if graph:
            if isinstance(graph, type):
                self.graph = graph({'weight': None, 'habituation_counter': 0, \
                        'radius': 0.0}, \
                    {'age': 0})
            else:
                self.graph = graph
        else:
            from .graph.array import ArrayGraph # default implementation
            self.graph = ArrayGraph({'weight': None, 'habituation_counter': 0, \
                    'radius': 0.0}, \
                {'age': 0})
        n1 = self.add_node(weight=w1)
        n2 = self.add_node(weight=w2)
        e  = self.connect(n1, n2)
        self.insertion_threshold = .95
        self.trust = 0 # in [0, 1]
        self.learning_rate = (.2, .006) # default values should be ok
        self.habituation_threshold = float('+inf') # useless, basically..
        self.habituation_initial = 1 # useless, basically..
        self.habituation_alpha = (1.05, 1.05) # should be greater than habituation_initial
        self.habituation_tau = (3.33, 14.33) # supposed to be time, but actually is to be
        # thought as a number of iterations; needs to be appropriately set
        self.edge_lifetime = 100 # may depend on the number of neighbors per node and
        # therefore on the dimensionality of the data;
        # also depends on the number of nodes, as lifetime reset may follow a Poisson distribution;
        # therefore, a floating point value below 1 will be interpreted as a fraction of the
        # number of nodes
        self.batch_size = 1000 # should be an order of magnitude or two below the total
        # sample size
        self.collapse_below = None
        self.knn = None
        self.topology = 'approximate density'

    def local_insertion_threshold(self, eta, node, *vargs):
        """
        """
        ##TODO: handle time more explicitly
        if self.knn:
            return max(self.insertion_threshold[0], \
                min(vargs[0], \
                    self.insertion_threshold[1]))
            # insertion_threshold are already radii (diameters / 2)
        else:
            return self.insertion_threshold

    def collapse_threshold(self, node1, node2):
        """
        """
        if self.knn:
            radius1 = self.get_node_attr(node1, 'radius')
            radius2 = self.get_node_attr(node2, 'radius')
            return self.collapse_below / self.insertion_threshold[0] * \
                max(radius1, radius2)
        else:
            return self.collapse_below

    def habituation_function(self, t, i=0):
        """
        Arguments:
            t (int or float or array): counter or time.
            i (int): proximity level (0=nearest, 1=second nearest).

        Returns:
            float or array: habituation.
        """
        return self.habituation_initial - \
            (1 - exp(-self.habituation_alpha[i] * t / self.habituation_tau[i])) / \
            (self.habituation_alpha[i])

    def habituation(self, node, i=0):
        """Returns the habituation of a node. Set ``i=0`` for the nearest node,
        ``i=1`` for a neighbor of the nearest node."""
        return self.habituation_function(float(self.get_node_attr(node, 'habituation_counter')), i)

    def plot_habituation(self):
        ##TODO: find a more general implementation for reading the habituations, and underlay
        # a histogram of the habituation counter in the graph
        import matplotlib.pyplot as plt
        import matplotlib.colors as clr
        import matplotlib.patches as patches
        if isinstance(self.graph, DictGraph):
            tmax = max(self.graph.nodes['habituation_counter'].values())
            tmax = round(float(tmax) / 100) * 100
        else:
            tmax = self.batch_size / 10
        t = np.arange(0, tmax)
        plt.plot(t, self.habituation_function(t, 1), 'c-')
        plt.plot(t, self.habituation_function(t, 0), 'b-')
        plt.xlim(0, tmax)
        plt.ylim(0, self.habituation_initial)
        plt.title('habituation function: alpha={}, tau={}'.format(self.habituation_alpha[0], \
                self.habituation_tau[0]))
        plt.ylabel('habituation')
        plt.xlabel('iteration number')

    def get_weight(self, node):
        return self.get_node_attr(node, 'weight')

    def set_weight(self, node, weight):
        self.set_node_attr(node, weight=weight)

    def increment_habituation(self, node):
        count = self.get_node_attr(node, 'habituation_counter') + 1
        self.set_node_attr(node, habituation_counter=count)
        return count

    def increment_age(self, edge):
        age = self.get_edge_attr(edge, 'age') + 1
        self.set_edge_attr(edge, age=age)
        return age

    def habituate(self, node):
        if self.edge_lifetime < 1:
            max_age = max(20, self.edge_lifetime * float(self.size))
        else:
            max_age = self.edge_lifetime
        self.increment_habituation(node)
        for edge, neighbor in list(self.iter_edges_from(node)):
            age = self.increment_age(edge)
            self.increment_habituation(neighbor) # increment before checking for age of the
            # corresponding edge, because `neighbor` may no exist afterwards; otherwise, to
            # increment after, it is necessary to check for existence of the node; should be
            # faster this way, due to low deletion rate
            if max_age < age:
                self.disconnect(node, neighbor, edge)
                if self.stands_alone(neighbor):
                    self.del_node(neighbor)
        if self.stands_alone(node):
            self.del_node(node)


    def batch_train(self, sample, eta_square=None, radius=None, grab=None, max_frames=None, **grab_kwargs):
        """This method grows the gas for a batch of data and implements the core GWR algorithm.
        :meth:`train` should be called instead."""
        if eta_square is None:
            eta_square = np.sum(sample * sample, axis=1)
        if radius is None:
            r = []
        errors = []
        for k in np.arange(0, sample.shape[0]):
            eta = sample[k]
            if radius is not None:
                r = [radius[k]]
            # find nearest and second nearest nodes
            dist2, index_to_node = self.square_distance('weight', eta, eta2=eta_square[k])
            i = np.argsort(dist2)
            dist2_min = dist2[i[0]]
            try:
                dist_min = sqrt(dist2_min)
            except ValueError:
                precision = int(np.dtype(dist2_min.dtype).str[-1])
                if (precision==8 and -1e-10 < dist2_min) or (precision==4 and -1e-3 < dist_min):
                    dist_min = 0
                    import warnings
                    warnings.warn('Rounding error: negative distance', RuntimeWarning)
                else:
                    print('square distance=', dist2_min, '   num. type=', dist2_min.dtype)
                    raise ValueError('Negative distance') from None
            nearest, second_nearest = index_to_node(i[:2])
            errors.append(dist_min)
            # test activity and habituation against thresholds
            activity = dist_min
            habituation = self.habituation(nearest)
            w = self.get_weight(nearest)
            activity_threshold = self.local_insertion_threshold(eta, w, *r)
            if isinstance(activity_threshold, tuple):
                assert not self.knn
                raise ValueError('insertion_threshold is defined as (lower, upper) bounds whereas knn is {}'.format(self.knn))
            if activity_threshold < activity and \
                habituation < self.habituation_threshold:
                # insert a new node and connect it with the two nearest nodes
                self.disconnect(nearest, second_nearest)
                l = .5 + self.trust * .5 # mixing coefficient in the range [.5, 1]
                new_node = self.add_node(weight=(1.0 - l) * w + l * eta)
                if self.knn:
                    self.set_node_attr(new_node, radius=activity_threshold)
                self.connect(new_node, nearest)
                self.connect(new_node, second_nearest)
            else:
                # move the nearest node and its neighbors towards the sample point
                self.connect(nearest, second_nearest)
                self.set_weight(nearest, w + self.learning_rate[0] * habituation * (eta - w))
                for i in self.iter_neighbors(nearest):
                    w = self.get_weight(i)
                    self.set_weight(i, w + self.learning_rate[1] * \
                        self.habituation(i, 1) * (eta - w))
            # update habituation counters
            self.habituate(nearest) # also habituates neighbors
            #
            if grab is not None:
                if max_frames is None or k < max_frames:
                    self.grab_frame(grab, **grab_kwargs)
        return errors

    def grab_frame(self, grab, axes=None, color='r', **kwargs):
        assert axes is not None
        xlim = axes.get_xlim()
        ylim = axes.get_ylim()
        # draw the graph
        lines = []
        for n1 in self.iter_nodes():
            w1 = self.get_weight(n1)
            for n2 in self.iter_neighbors(n1):
                w2 = self.get_weight(n2)
                lines.append(axes.plot([w1[0], w2[0]], [w1[1], w2[1]], color+'-', linewidth=1))
        #
        axes.set_xlim(xlim)
        axes.set_ylim(ylim)
        axes.set_axis_off()
        # grab
        grab.grab_frame()
        # clear the graph
        while lines:
            lines.pop().pop()

    def grab_batch_init(self, grab, sample, i, batch, axes=None, max_frames=None, **kwargs):
        assert axes is not None
        n = sample.shape[0]
        _batch = np.zeros(n, dtype=bool)
        _batch[batch] = True
        axes.clear()
        axes.plot(sample[~_batch,0], sample[~_batch,1], '.', color='lightgray', markersize=1)
        axes.plot(sample[_batch,0], sample[_batch,1], 'k.', markersize=1)
        kwargs = dict(axes=axes)
        if max_frames is not None:
            kwargs['max_frames'] = max_frames - (i-1)*self.batch_size
        self.grab_frame(grab, **kwargs)
        return kwargs

    def grab_completion(self, grab, sample, axes=None, **kwargs):
        assert axes is not None
        axes.clear()
        axes.plot(sample[:,0], sample[:,1], '.', color='lightgray', markersize=1)
        self.grab_frame(grab, axes, 'k')

    def train(self, sample, pass_count=None, residual_max=None, error_count_tol=1e-6, \
        min_growth=None, collapse_tol=None, stopping_criterion=2, verbose=False, \
        plot=False, grab=None, **kwargs):
        """
        Grow the gas.

        :meth:`train` splits the sample into batches, successively calls :meth:`batch_train`
        on these batches of data, collapses the gas if necessary and stops if stopping criteria
        are met.

        The input arguments that define the stopping criteria are:

        Arguments:

            pass_count (pair of floats or None):
                (min, max) numbers of passes of the sample the algorithm should/can run.
                Any bound can be set to ``None``.
            residual_max (float):
                a residual is calculated for each sample point, before fitting the gas
                towards it, and `residual_max` is a threshold above which a residual is
                regarded as an 'error'. This parameter works in combination with
                `error_count_tol`.
            error_count_tol (float):
                maximum proportion of 'errors' (number of errors in a batch over the
                size of the batch). If this maximum is exceeded, then the :meth:`train`
                samples another batch and keeps on training the gas. Otherwise, the next
                criteria apply.
            min_growth (float):
                minimum relative increase in the number of nodes from an iteration to
                the next.
            collapse_tol (float):
                maximum allowed ratio of the number of collapsed nodes over the total
                number of nodes.
            stopping_criterion (int, deprecated):

                1. performs a linear regression for the pre-fitting residual across time
                   and 'tests' whether the trend is not negative.
                2. stops if the average residual for the current batch is greater than
                   that of the previous batch.


        """
        ## TODO: clarify the code
        if isinstance(sample, tuple):
            locations, displacements = sample
            sample = locations
        elif self.knn and self.topology == 'displacement length':
            raise ValueError('no displacement information found')
        n = sample.shape[0]
        l = 0 # node count
        residuals = []
        if stopping_criterion == 1:
            import statsmodels.api as sm
        # if stopping_criterion is 2
        residual_mean = float('+inf')
        #
        if pass_count:
            p = .95
            #k1 = log(1 - p) / log(1 - 1 / float(n)) # sample size for each point to have probability p of being chosen once
            k1 = n
        if residual_max:
            if error_count_tol < 1:
                error_count_tol = ceil(error_count_tol * float(n))
        # local radius
        if self.knn:
            if not isinstance(self.insertion_threshold, tuple):
                self.insertion_threshold = (self.insertion_threshold, \
                    self.insertion_threshold * 4)
            eta_square = np.sum(sample * sample, axis=1)
            if self.topology == 'approximate density':
                radius = self.boxed_radius(sample, self.knn, self.insertion_threshold[0], \
                        self.insertion_threshold[1], verbose, plot)
            elif self.topology == 'displacement length':
                radius = self.boxed_average(sample, displacements, self.knn, self.insertion_threshold[0], \
                        self.insertion_threshold[1], verbose, plot)
            else:
                raise ValueError("topology=='{}' not supported".format(self.topology))
        # grab
        batch_kwargs = {}
        if grab is not None:
            batch_kwargs['grab'] = grab
            if sample.shape[1] != 2:
                raise ValueError('cannot grab data that are not 2D')
            max_frames = kwargs.get('max_frames', None)
            max_batches = kwargs.get('max_batches', None)
        # loop
        t = []
        i = 0
        do = True
        while do:
            i += 1
            if verbose:
                t0 = time.time()
            batch = np.random.choice(n, size=self.batch_size)
            # grab
            if grab is not None:
                if max_batches is None or i <= max_batches:
                    batch_kwargs.update(self.grab_batch_init(grab, sample, i, batch, **kwargs))
                    if 1 < i and \
                            (max_frames is None or \
                                (max_batches is not None and \
                                    max_frames <= (i-1)*self.batch_size)):
                        batch_kwargs = {}
                else:
                    batch_kwargs = {}
            #
            if self.knn:
                r = self.batch_train(sample[batch], eta_square[batch], radius[batch], **batch_kwargs)
            else:
                r = self.batch_train(sample[batch], **batch_kwargs)
            residuals += r
            l_prev = l
            l = self.size
            txt = l
            if self.collapse_below:
                self.collapse()
                dl = l - self.size
                if verbose and 0 < dl:
                    txt = '{} (-{})'.format(l, dl)
            if verbose:
                if i == 1:
                    if self.collapse_below:
                        print("\t#nodes (-#collapsed)")
                    else:   print("\t#nodes")
                ti = time.time() - t0
                t.append(ti)
                print("{}\t{}\tElapsed: {:.0f} ms".format(i, txt, ti * 1e3))
            # enforce some stopping or continuing conditions
            if pass_count:
                k = i * self.batch_size
                if pass_count[0] and k < pass_count[0] * k1:
                    #print(('pass_count', k, pass_count[0] * k1))
                    continue
                elif pass_count[1] and pass_count[1] * k1 <= k:
                    if verbose:
                        print('stopping criterion: upper bound for `pass_count` reached')
                    break # do = False
            if residual_max:
                error_count = len([ residual for residual in r
                        if residual_max < residual ])
                if error_count_tol < error_count:
                    #print(('error_count_tol', error_count_tol, error_count))
                    continue
            if min_growth is not None:
                growth = float(l - l_prev) / float(l_prev)
                if growth < min_growth:
                    if verbose:
                        print('stopping criterion: relative growth: {:.0f}%'.format(growth * 100))
                    break
            if self.collapse_below and collapse_tol:
                collapse_rel = float(dl) / float(l)
                if collapse_tol < collapse_rel:
                    if verbose:
                        print('stopping criterion: relative collapse: {:.0f}%'.format(collapse_rel * 100))
                    break
            # exclusive criteria
            if stopping_criterion == 1:
                regression = sm.OLS(np.array(r), \
                        sm.add_constant(np.linspace(0,1,len(r)))\
                    ).fit()
                if 0 < regression.params[1]:
                    fit = 0
                else:
                    fit = 1/(1+exp((.1-regression.pvalues[1])/.01)) # invert p-value
                do = tolerance < fit
            elif stopping_criterion == 2:
                residual_prev = residual_mean
                #_, r = self.eval(sample[np.random.choice(n, size=self.validation_batch_size),:])
                residual_mean = np.mean(r)
                residual_std  = np.std(r)
                do = (residual_mean - residual_prev) / residual_std < 0
        if verbose:
            t = np.asarray(t)
            print('Elapsed:  mean: {:.0f} ms  std: {:.0f} ms'.format(np.mean(t) * 1e3, \
                                    np.std(t) * 1e3))
        if grab is not None and (max_frames is None or max_frames < sample.shape[0]):
            self.grab_completion(grab, sample, **kwargs)
        return residuals

    def boxed_radius(self, sample, knn, rmin, rmax, verbose=False, plot=False):
        #plot = True
        if plot:
            import matplotlib.pyplot as plt
        d = sample.shape[1]
        d = int(d) # PY2
        #if d == 2:
        #       return self.boxed_radius2d(sample, knn, rmin, rmax, plot) # faster
        if verbose:
            t0 = 0
            t = time.time()
        # bounding box(es)
        sample = np.asarray(sample)
        smin = sample.min(axis=0)
        smax = sample.max(axis=0)
        unit = rmin / sqrt(float(d)) # min radius will be the diagonal of the hypercube
        # volume ratio of the unit hypercube (1) and the inscribed d-ball
        dim_penalty = 1 / (pi ** (float(d) / 2.0) / gamma(float(d) / 2.0 + 1.0) * 0.5 ** d)
        max_n_units = ceil(2.0 * rmax / unit * dim_penalty) # max number of "circles" where to look for neighbors
        bmin, bmax = smin, smax # multiple names because of copying/pasting
        dim = np.ceil((bmax - bmin) / unit) - 1
        adjusted_bmax = bmin + dim * unit # bmin and adjusted_bmax are the lower vertices (origins) of the end cubes
        dim = [ int(_d) + 1 for _d in dim ] # _d for PY2
        # partition
        cell = dict()
        cell_indices = dict()
        count = np.zeros(dim, dtype=int)
        grid = Dichotomy(sample, base_edge=unit, origin=bmin, max_level=0)
        grid.split()
        for j in grid.cell:
            i, _, k = grid.cell[j]
            ids, pts = grid.subset[k]
            if ids.size:
                i = tuple([ int(_n) for _n in np.round((i - bmin) / unit) ])
                cell[i] = pts
                if ids.size < knn + 1:
                    cell_indices[i] = ids
                try:
                    count[i] = ids.size
                except IndexError:
                    print((count.shape, i))
                    raise
        # counts are density estimate; check how good/poor they are
        if d == 2 and plot:
            if verbose:
                t0 += time.time() - t
            plt.imshow(count.T)
            plt.gca().invert_yaxis()
            plt.show()
            if verbose:
                t = time.time()
        # count_threshold is useful for saving computation time,
        # skipping poorly populated cells with populated neighbor cells.
        # lower threshold means higher skip frequency (more approximations)
        count_threshold = np.zeros_like(count)
        for i in range(d):
            a = np.swapaxes(count_threshold, 0, i) # view (numpy >= 1.10)
            b = np.swapaxes(count, 0, i) # view
            a[:-1,...] += b[1::,...]
            a[1::,...] += b[:-1,...]
        count_threshold = (knn + 1) - count_threshold / (d * 2) # average number of points in neighbor cells
        # count_threshold should depend on space dimension
        count_threshold *= 2 # hard-coded, empirical, actually not even adjusted
        # adjust the requested number of neighbors (k)
        k = int(round(float(knn) * dim_penalty)) # request more neighbors in the cube so that the inscribed ball should contain knn as expected (hypothesis: that points are homogeneously distributed
        k *= 2 # gwr actually works better with higher k; hard-coded factor; empirical
        scale = {}
        n = sample.shape[0]
        radius = rmin * np.ones(n, dtype=sample.dtype)
        for i in cell_indices:
            if count_threshold[i] < count[i]:
                continue # consider that local radius are minimal (saves time)
            m = 0
            while True: # do..while
                m += 1
                I = [ np.arange(max(0, _k - m), min(_k + m + 1, _d)) \
                    for _k, _d in zip(i, dim) ] # _k, _d for PY2
                I = np.meshgrid(*I, indexing='ij')
                I = np.column_stack([ _i.flatten() for _i in I ]) # _i for PY2
                p = np.sum(count[tuple(I.T)])
                if not (p < k + 1 and m < max_n_units):
                    break
            scale[i] = m
            j = cell_indices[i]
            if p < k + 1:
                # `m` has reached `max_n_units`
                radius[j] = rmax
                continue
            X = []
            for _i in I:
                try:
                    _x = cell[tuple(_i)]
                except KeyError:
                    pass
                else:
                    X.append(_x)
            #X = [ cell[tuple(_i)] for _i in I if tuple(_i) in cell ] # _i for PY2
            X = np.concatenate(X, axis=0)
            x = cell[i]
            # dist with SciPy
            D = cdist(X, x)
            D.sort(axis=0)
            r = D[k]
            # and without SciPy (better for bigger matrices)
            #D = np.dot(X, x.T)
            #x2 = np.sum(x * x, axis=1, keepdims=True)
            #D -= 0.5 * x2.T
            #X2 = np.sum(X * X, axis=1, keepdims=True)
            #D -= 0.5 * X2
            #D.sort(axis=0)
            #r = D[-(k+1)]
            #r = np.sqrt(-2.0 * r)
            #r[np.isnan(r)] = 0.0 # numeric precision errors => negative square distances => NaN distances
            #
            del D
            radius[j] = r
        if verbose:
            print('estimating density: elapsed time {:d} ms'.format(int(round((t0 + time.time() - t) * 1000)))) # int for PY2
        # plot local radius
        #plot = True
        if d == 2 and plot:
            color = (radius - rmin) / (rmax - rmin)
            r = self.collapse_below * 0.5 # collapse_below defines a diameter
            cmap = plt.get_cmap('RdPu')
            fig = plt.figure()
            ax = fig.add_subplot(111, aspect='equal')
            for x in np.arange(smin[0], smax[0], unit):
                plt.plot([x, x], [smin[1], smax[1]], 'y-')
            for y in np.arange(smin[1], smax[1], unit):
                plt.plot([smin[0], smax[0]], [y, y], 'y-')
            for i in range(n):
                c = color[i]
                if rmax < radius[i]:
                    ax.add_artist(patches.Circle(sample[i], r * 4, \
                        color=cmap(0.0), alpha=0.1))
                elif radius[i] < rmin - np.spacing(1):
                    ax.add_artist(patches.Circle(sample[i], r * 0.25, \
                        color=cmap(1.0)))
                elif rmin + np.spacing(1) < radius[i]:
                    ax.add_artist(patches.Circle(sample[i], r, \
                        color=cmap(1.0 - c)))
            plt.plot(sample[:,0], sample[:,1], 'k+')
            for i, x in cell.items():
                center = smin + (np.asarray(list(i)) + 0.5) * unit
                count = x.shape[0]
                if count < knn + 1:
                    color = 'm'
                    txt = "{:d}\n{:d}".format(count, scale.get(i, 0))
                else:
                    color = 'c'
                    txt = str(count)
                plt.text(center[0], center[1], txt, color=color, \
                    horizontalalignment='center', verticalalignment='center')
            try:
                plt.show()
            except AttributeError: # window has been closed; go on
                pass
        return radius


    def boxed_radius2d(self, sample, knn, rmin, rmax, plot=False):
        # deprecated: too many approximations here
        if plot:
            import matplotlib.pyplot as plt
        t = time.time()
        n = sample.shape[0]
        radius = rmin * np.ones(n, dtype=sample.dtype)
        # bounding box(es)
        smin = sample.min(axis=0)
        smax = sample.max(axis=0)
        unit = rmin * sqrt(2.0)
        max_n_units = ceil(2.0 * rmax / unit)
        ix = np.arange(n)
        # grid
        bounding_box = [(smin, smax), \
            (smin + unit * 0.25, smax - unit * 0.75), \
            (smin + unit * 0.5,  smax - unit * 0.5 ), \
            (smin + unit * 0.75, smax - unit * 0.25)]
        for bmin, bmax in bounding_box:
            bmax_ = np.ceil(bmax / unit) * unit + np.spacing(1)
            x_grid = np.arange(bmin[0], bmax_[0], unit)
            y_grid = np.arange(bmin[1], bmax_[1], unit)
            # partition strategy
            if x_grid.size <= y_grid.size:
                grid0 = x_grid
                grid1 = y_grid
                x0 = sample[:, 0]
                x1 = sample[:, 1]
            else:
                grid0 = y_grid
                grid1 = x_grid
                x0 = sample[:, 1]
                x1 = sample[:, 0]
            n0 = grid0.size - 1
            n1 = grid1.size - 1
            # partition
            cell = dict()
            icell = dict()
            undefined = []
            lower0 = np.ones(n, dtype=np.bool)
            for i in range(n0):
                upper0 = x0 < grid0[i+1]
                mask0 = np.logical_and(lower0, upper0)
                xi = x1[mask0]
                ix0 = ix[mask0]
                lower0 = np.logical_not(upper0)
                ni = xi.shape[0]
                lower1 = np.ones(ni, dtype=np.bool)
                for j in range(n1):
                    upper1 = xi < grid1[j+1]
                    ix1 = ix0[np.logical_and(lower1, upper1)]
                    xj = sample[ix1]
                    cell[(i,j)] = xj
                    if 0 < xj.size and xj.shape[0] < knn:
                        undefined.append((i,j))
                        icell[(i,j)] = ix1
                    lower1 = np.logical_not(upper1)
            #print(len(undefined))
            # compute radius where point density is low
            for i, j in undefined:
                nij = 0
                m = 0
                while nij < knn:
                    m += 1
                    Xij = [ cell[(k,l)] for k in range(i-m,i+m) for l in range(j-m,j+m) \
                        if 0 <= k and k < n0 and 0 <= l and l < n1 ]
                    nij = sum([ x.shape[0] for x in Xij ])
                Xij = np.concatenate(Xij, axis=0)
                xij = cell[(i,j)]
                D = np.dot(xij, Xij.T)
                x2 = np.sum(xij * xij, axis=1)
                x2.shape = (x2.size, 1)
                D -= 0.5 * x2
                X2 = np.sum(Xij * Xij, axis=1)
                X2.shape = (1, X2.size)
                D -= 0.5 * X2
                D.sort()
                r = D[:, (nij - 1) - knn]
                del D
                r = np.sqrt(-2.0 * r)
                r[np.isnan(r)] = 0.0
                #if any(r < rmin):
                #       print(r)
                radius[icell[(i,j)]] = np.maximum(radius[icell[(i,j)]], r)
        print('elapsed time {:d} ms'.format(round((time.time() - t) * 1000)))
        # plot local radius
        if plot:
            color = np.minimum((radius - rmin) / (rmax - rmin), 1.0) * 0.9
            cmap = plt.get_cmap('RdPu')
            fig = plt.figure()
            ax = fig.add_subplot(111, aspect='equal')
            for i in range(n):
                c = color[i]
                if np.spacing(1) < c:
                    ax.add_artist(patches.Circle(sample[i], radius[i], \
                        color=cmap(1 - c), alpha=0.2))
            plt.plot(sample[:,0], sample[:,1], 'k+')
            plt.show()
        return radius

    def exact_radius(self, sample, eta_square, knn, step, *vargs):
        # deprecated: extremely slow
        n = sample.shape[0]
        w = sample.astype(np.float32)
        w2 = eta_square.astype(np.float32) * 0.5
        w2.shape = (w2.size, 1)
        D = []
        for i in range(0, n, step):
            j = range(i, min(i + step, n))
            t = time.time()
            d = np.dot(w[j], w.T)
            d -= w2[j]
            d -= w2.T
            print("Elapsed: {:.0f} ms".format((time.time() - t) * 1e3))
            t = time.time()
            d.sort() # along last axis if faster
            print("Elapsed: {:.0f} ms".format((time.time() - t) * 1e3))
            d = d[:, (n - 1) - (self.knn + 1)].astype(sample.dtype)
            d = np.sqrt(-2.0 * d)
            D.append(d)
        radius = np.concatenate(D)
        return radius


    def boxed_average(self, locations, displacements, knn, rmin, rmax, verbose=False, plot=False):
        #plot = True
        if plot:
            import matplotlib.pyplot as plt
        if not locations.shape == displacements.shape:
            raise ValueError('locations and displacements do not match in shape')
        d = locations.shape[1]
        d = int(d) # PY2
        if verbose:
            t0 = 0
            t = time.time()
        # bounding box(es)
        sample = np.asarray(locations)
        displacements = np.asarray(displacements)
        smin = sample.min(axis=0)
        smax = sample.max(axis=0)
        unit = rmin / sqrt(float(d)) # min radius will be the diagonal of the hypercube
        # volume ratio of the unit hypercube (1) and the inscribed d-ball
        dim_penalty = 1 / (pi ** (float(d) / 2.0) / gamma(float(d) / 2.0 + 1.0) * 0.5 ** d)
        max_n_units = ceil(2.0 * rmax / unit * dim_penalty) # max number of "circles" where to look for neighbors
        bmin, bmax = smin, smax # multiple names because of copying/pasting
        dim = np.ceil((bmax - bmin) / unit) - 1
        adjusted_bmax = bmin + dim * unit # bmin and adjusted_bmax are the lower vertices (origins) of the end cubes
        dim = [ int(_d) + 1 for _d in dim ] # _d for PY2
        # partition
        cell = dict()
        cell_indices = dict()
        count = np.zeros(dim, dtype=int)
        grid = Dichotomy(sample, base_edge=unit, origin=bmin, max_level=0)
        grid.split()
        for j in grid.cell:
            i, _, k = grid.cell[j]
            ids, pts = grid.subset[k]
            if ids.size:
                i = tuple([ int(_n) for _n in np.round((i - bmin) / unit) ])
                cell[i] = pts
                if ids.size < knn + 1:
                    cell_indices[i] = ids
                try:
                    count[i] = ids.size
                except IndexError:
                    print((count.shape, i))
                    raise
        # counts are density estimate; check how good/poor they are
        if d == 2 and plot:
            if verbose:
                t0 += time.time() - t
            plt.imshow(count.T)
            plt.gca().invert_yaxis()
            plt.show()
            if verbose:
                t = time.time()
        # count_threshold is useful for saving computation time,
        # skipping poorly populated cells with populated neighbor cells.
        # lower threshold means higher skip frequency (more approximations)
        count_threshold = np.zeros_like(count)
        for i in range(d):
            a = np.swapaxes(count_threshold, 0, i) # view (numpy >= 1.10)
            b = np.swapaxes(count, 0, i) # view
            a[:-1,...] += b[1::,...]
            a[1::,...] += b[:-1,...]
        count_threshold = (knn + 1) - count_threshold / (d * 2) # average number of points in neighbor cells
        # count_threshold should depend on space dimension
        count_threshold *= 2 # hard-coded, empirical, actually not even adjusted
        # adjust the requested number of neighbors (k)
        k = int(round(float(knn) * dim_penalty)) # request more neighbors in the cube so that the inscribed ball should contain knn as expected (hypothesis: points are homogeneously distributed)
        k *= 2 # gwr actually works better with higher k; hard-coded factor; empirical
        scale = {}
        n = sample.shape[0]
        dr = displacements
        displacement_length = np.sqrt(np.sum(dr*dr, axis=1))
        avg_length = np.zeros(n, dtype=sample.dtype)
        for i in cell_indices:
            j = cell_indices[i]
            if count_threshold[i] < count[i]:
                avg_length[j] = np.median(displacement_length[j], keepdims=True)
                continue # consider more measurements than wished
            m = 0
            while True:
                m += 1
                I = [ np.arange(max(0, _k - m), min(_k + m + 1, _d)) \
                    for _k, _d in zip(i, dim) ] # _k, _d for PY2
                I = np.meshgrid(*I, indexing='ij')
                I = np.column_stack([ _i.flatten() for _i in I ]) # _i for PY2
                p = np.sum(count[tuple(I.T)])
                if not (p < k + 1 and m < max_n_units):
                    break
            scale[i] = m
            if p < k + 1:
                # `m` has reached `max_n_units`; consider less measurements than wished
                avg_length[j] = np.median(displacement_length[j], keepdims=True)
                continue
            X = []
            L = []
            for _i in I:
                try:
                    _len = displacement_length[cell_indices[tuple(_i)]]
                    _x = cell[tuple(_i)]
                except KeyError:
                    pass
                else:
                    L.append(_len)
                    X.append(_x)
            #X = [ cell[tuple(_i)] for _i in I if tuple(_i) in cell ] # _i for PY2
            X = np.concatenate(X, axis=0)
            L = np.concatenate(L, axis=0)
            x = cell[i]
            # dist with SciPy
            D = cdist(X, x)
            I = np.argsort(D, axis=0)[:k]
            # and without SciPy (better for bigger matrices)
            #D = np.dot(X, x.T)
            #x2 = np.sum(x * x, axis=1, keepdims=True)
            #D -= 0.5 * x2.T
            #X2 = np.sum(X * X, axis=1, keepdims=True)
            #D -= 0.5 * X2
            #D.sort(axis=0)
            #r = D[-(k+1)]
            #r = np.sqrt(-2.0 * r)
            #r[np.isnan(r)] = 0.0 # numeric precision errors => negative square distances => NaN distances
            #
            del D
            avg_length[j] = np.median(L[I], axis=0)
        if verbose:
            print('estimating density: elapsed time {:d} ms'.format(int(round((t0 + time.time() - t) * 1000)))) # int for PY2
        # plot local radius
        #plot = True
        if d == 2 and plot:
            color = (avg_length - rmin) / (rmax - rmin)
            r = self.collapse_below * 0.5 # collapse_below defines a diameter
            cmap = plt.get_cmap('RdPu')
            fig = plt.figure()
            ax = fig.add_subplot(111, aspect='equal')
            for x in np.arange(smin[0], smax[0], unit):
                plt.plot([x, x], [smin[1], smax[1]], 'y-')
            for y in np.arange(smin[1], smax[1], unit):
                plt.plot([smin[0], smax[0]], [y, y], 'y-')
            for i in range(n):
                c = color[i]
                if rmax < avg_length[i]:
                    ax.add_artist(patches.Circle(sample[i], r * 4, \
                        color=cmap(0.0), alpha=0.1))
                elif avg_length[i] < rmin - np.spacing(1):
                    ax.add_artist(patches.Circle(sample[i], r * 0.25, \
                        color=cmap(1.0)))
                elif rmin + np.spacing(1) < avg_length[i]:
                    ax.add_artist(patches.Circle(sample[i], r, \
                        color=cmap(1.0 - c)))
            plt.plot(sample[:,0], sample[:,1], 'k+')
            for i, x in cell.items():
                center = smin + (np.asarray(list(i)) + 0.5) * unit
                count = x.shape[0]
                if count < knn + 1:
                    color = 'm'
                    txt = "{:d}\n{:d}".format(count, scale.get(i, 0))
                else:
                    color = 'c'
                    txt = str(count)
                plt.text(center[0], center[1], txt, color=color, \
                    horizontalalignment='center', verticalalignment='center')
            try:
                plt.show()
            except AttributeError: # window has been closed; go on
                pass
        return avg_length


    def collapse(self):
        for n in list(self.iter_nodes()):
            if self.has_node(n):
                neighbors = [ neighbor for neighbor in self.iter_neighbors(n)
                            if n < neighbor ]
                if neighbors:
                    dist = la.norm(self.get_weight(n) - \
                        np.vstack([ self.get_weight(n)
                            for n in neighbors ]), axis=1)
                    k = np.argmin(dist)
                    neighbor = neighbors[k]
                    if dist[k] < self.collapse_threshold(n, neighbor):
                        #print((n, neighbor))
                        self.collapse_nodes(n, neighbor)

    def collapse_nodes(self, n1, n2):
        self.disconnect(n1, n2)
        w1 = self.get_weight(n1)
        w2 = self.get_weight(n2)
        self.set_weight(n1, (w1 + w2) / 2)
        for e2, n in list(self.iter_edges_from(n2)):
            a2 = self.get_edge_attr(e2, 'age')
            e1 = self.find_edge(n1, n)
            if e1 is None:
                self.connect(n1, n, age=a2)
            else:
                a1 = self.get_edge_attr(e1, 'age')
                self.set_edge_attr(e1, age=min(a1, a2))
            #self.disconnect(n2, n, e2)
        self.del_node(n2)





"""

Runtime configuration.

To globally disable *tqdm* console animations, or yes-no questions, for example at file overwrite:

.. code:python

    from tramway.core import rc
    rc.__user_interaction__ = False

For optional dependencies, for example:

.. code:python

    try:
        from tqdm import tqdm
    except ImportError:
        pass
    else:
    rc.__available_packages__.add('tqdm')

and then:

.. code:python

    if rc.__has_package__('tqdm'):
        ...

"""

import warnings
__log__ = lambda msg: warnings.warn(msg, RuntimeWarning) # could call the logging module instead


__user_interaction__ = True


__available_packages__ = set()
__reported_missing_packages__ = set()
def __has_package__(pkg, report_if_missing=True):
    if pkg in __available_packages__:
        return True
    elif report_if_missing:
        if pkg not in __reported_missing_packages__:
            __reported_missing_packages__.add(pkg)
            __log__("package '{}' is missing".format(pkg))
    return False


# example usage:
#try:
#    import tqdm
#except ImportError:
#    pass
#else:
#    __available_packages__.add('tqdm')


__all__ = ['__user_interaction__', '__available_packages__', '__has_package__']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute.abc import Attribute, abstractmethod

class ROI(Attribute):
    """
    Abstract base class for the :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute
    of an :class:`~tramway.analyzer.RWAnalyzer` object.
    """
    @abstractmethod
    def __iter__(self):
        return self.as_individual_roi()
    @abstractmethod
    def as_individual_roi(self, index=None, collection=None):
        """ 
        Generator function; iterates over the individual ROI.

        `index` is the index in the collection and `collection`
        should be defined if multiple collections of ROI are defined.

        .. note::

            Iterating over the individual ROI may be useful for visualization;
            favor :meth:`~tramway.analyzer.roi.ROI.as_support_regions` for
            data processing.

        """
        pass
    @abstractmethod
    def as_support_regions(self, index=None):
        """
        Generator function; iterates over the support regions.

        Support regions are the regions to be processed.
        An individual ROI operates as a window on a support region.
        """
        pass




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from ..roi import DecentralizedROIManager
import sys
import traceback


class PipelineStage(object):
    """
    Wrapper for callables to be run on calling :meth:`Pipeline.run`.

    `requires_mutability` defaults to :const:`False`.

    `update_existing_rwa_files` defaults to :const:`False` as of version *5.2*.
    """
    __slots__ = ('_run','_granularity','_mutability','options','update_existing_rwa_files')
    def __init__(self, run, granularity=None, requires_mutability=None,
            update_existing_rwa_files=None, **options):
        if isinstance(run, PipelineStage):
            self._run = run._run
            self._granularity = run._granularity if granularity is None else granularity
            self._mutability = run._mutability if requires_mutability is None else requires_mutability
            self.options = dict(run.options)
            self.options.update(options)
            self.update_existing_rwa_files = run.update_existing_rwa_files \
                    if update_existing_rwa_files is None else update_existing_rwa_files
        else:
            self._run = run
            self._granularity = granularity
            self._mutability = False if requires_mutability is None else requires_mutability
            self.options = options
            self.update_existing_rwa_files = update_existing_rwa_files
    @property
    def granularity(self):
        """
        *str*: See :meth:`Pipeline.append_stage`
        """
        return self._granularity
    @property
    def requires_mutability(self):
        """
        *bool*: See :meth:`Pipeline.append_stage`
        """
        return self._mutability
    @property
    def name(self):
        """
        *str*: Callable's name
        """
        return self._run.__name__
    def __call__(self, *args, **kwargs):
        return self._run(*args, **kwargs)


class Pipeline(AnalyzerNode):
    """
    :attr:`~tramway.analyzer.RWAnalyzer.pipeline` attribute of an
    :class:`~tramway.analyzer.RWAnalyzer` object.

    The main methods are :meth:`append_stage` and :meth:`run`.
    """
    __slots__ = ('_stage',)
    def __init__(self, *args, **kwargs):
        AnalyzerNode.__init__(self, *args, **kwargs)
        self._stage = []
    def __nonzero__(self):
        return bool(self._stage)
    def __len__(self):
        return len(self._stage)
    @property
    def analyzer(self):
        return self._parent
    @property
    def spt_data(self):
        return self._parent.spt_data
    @property
    def roi(self):
        return self._parent.roi
    @property
    def time(self):
        return self._parent.time
    @property
    def tesseller(self):
        return self._parent.tesseller
    @property
    def sampler(self):
        return self._parent.sampler
    @property
    def mapper(self):
        return self._parent.mapper
    @property
    def env(self):
        return self._parent.env
    def reset(self):
        """
        Empties the pipeline processing chain.
        """
        self._stage = []
    def append_stage(self, stage, granularity=None, requires_mutability=None, **options):
        """
        Appends a pipeline stage to the processing chain.

        Arguments:

            stage (callable): function that takes an :class:`~tramway.analyzer.RWAnalyzer` object
                as unique input argument.

            granularity (str): smallest data item `stage` can independently process;
                any of :const:`'coarsest'` or equivalently :const:`'full dataset'`,
                :const:`'source'` or equivalently :const:`'spt data'`, :const:`'data source'`
                or :const:`'spt data source'`, :const:`'roi'` or equivalently
                :const:`'region of interest'`,
                :const:`'time'` or equivalently :const:`'segment'` or :const:`'time segment'`
                (case-insensitive).

            requires_mutability (bool): callable object `stage` alters input argument `self`.
                Stages with this argument set to :const:`True` are always run as dependencies.
                Default is :const:`False`.

            update_existing_rwa_files (bool): see also :class:`PipelineStage`.

        """
        self._stage.append(PipelineStage(stage, granularity, requires_mutability, **options))
    def early_setup(self, **kwargs):
        """
        Sets the `submit_side`/`worker_side` properties of the :attr:`~tramway.analyzer.RWAnalyzer.env`
        attribute.

        :meth:`early_setup` can be called once :attr:`~tramway.analyzer.RWAnalyzer.env` is initialized
        and before :meth:`run` is called.
        This allows to run some conditional code, specifically on the submit side or on the worker
        side.

        This method returns silently if :attr:`~tramway.analyzer.RWAnalyzer.env` is not initialized
        and both the :attr:`..env.EnvironmentInitializer.submit_side` and
        :attr:`..env.EnvironmentInitializer.worker_side` properties are :const:`False`.
        """
        if self.env.initialized:
            self.env.early_setup(*sys.argv, **kwargs)
    def run(self):
        """
        Sequentially runs the different stages of the pipeline.
        """
        if self.env.initialized:
            try:
                self.env.setup(*sys.argv)
                self.logger.info('setup complete')
                if self.env.worker_side and not self.env.submit_side:
                    # a single stage can apply
                    stage_index = self.env.selectors.get('stage_index', 0)
                    if isinstance(stage_index, (tuple,list)):
                        for i in stage_index[:-1]:
                            stage = self._stage[i]
                            stage(self)
                        stage_index = stage_index[-1]
                    stage = self._stage[stage_index]
                    # alter the iterators for spt_data
                    self.analyzer.spt_data.self_update(self.env.spt_data_selector)
                    # alter the iterators for roi
                    if isinstance(self.roi, DecentralizedROIManager):
                        for f in self.spt_data:
                            f.roi.self_update(self.env.roi_selector)
                    elif not isinstance(self.roi, Initializer):
                        self.analyzer.roi.self_update(self.env.roi_selector)
                    # alter the iterators for time
                    if not isinstance(self.time, Initializer):
                        self.analyzer.time.self_update(self.env.time_selector)
                    self.logger.info('stage {:d} ready'.format(stage_index))
                    try:
                        stage(self)
                    except:
                        self.logger.error('stage {:d} failed with t'.format(stage_index)+traceback.format_exc()[1:-1])
                        raise
                    else:
                        self.logger.info('stage {:d} done'.format(stage_index))
                    #
                    #self.env.save_analyses(self.spt_data)
                else:
                    assert self.env.submit_side
                    if self.env.dispatch():
                        self.logger.info('initial dispatch done')
                    stack = []
                    permanent_stack = []
                    for s, stage in enumerate(self._stage):
                        granularity = '' if stage.granularity is None \
                                else stage.granularity.lower().replace('-',' ').replace('_',' ')
                        if stage.requires_mutability:
                            if not granularity or granularity in ('coarsest','full dataset'):
                                # run locally
                                self.logger.debug('[submit] stage {:d} ready'.format(s))
                                stage(self)
                                self.logger.debug('[submit] stage {:d} done'.format(s))
                                # make the next dispatched stages run this stage again
                                permanent_stack.append(s)
                                continue
                            elif stage.requires_mutability:
                                raise NotImplementedError('cannot make a dispatched job modify the local analyzer')
                        if self.env.dispatch(stage_index=s, stage_options=stage.options):
                            self.logger.info('stage {:d} dispatched'.format(s))
                        if stack or permanent_stack:
                            s = sorted(permanent_stack+stack+[s])
                            stack = []
                        if not granularity or granularity in ('coarsest','full dataset'):
                            self.env.make_job(stage_index=s)
                        else:
                            try:
                                alias = all([ bool(f.alias) for f in self._eldest_parent.spt_data ])
                            except AttributeError:
                                alias = False
                            for f in self.spt_data:
                                source = f.alias if alias else f.source
                                if source is None:
                                    if 1<len(self.spt_data):
                                        raise NotImplementedError('undefined source identifiers')
                                elif self.env.dispatch(source=source):
                                    self.logger.info('source "{}" dispatched'.format(source))
                                if granularity.endswith('source') or granularity.startswith('spt data'):
                                    self.env.make_job(stage_index=s, source=source)
                                elif granularity in ('roi','region of interest'):
                                    for i, _ in f.roi.as_support_regions(return_index=True):
                                        self.env.make_job(stage_index=s, source=source, region_index=i)
                                elif granularity in ('time', 'segment', 'time segment'):
                                    for i, r in f.roi.as_support_regions(return_index=True):
                                        try:
                                            w = r.get_sampling()
                                        except ValueError:
                                            raise NotImplementedError('cannot iterate on multiple sampling per ROI yet') from None
                                        except KeyError:
                                            raise NotImplementedError('cannot autoload the sampling stage; please load the sampling in a separate stage with requires_mutability=True') from None
                                        for j, _ in self.time.as_time_segments(w, return_index=True, return_times=False):
                                            self.env.make_job(stage_index=s, source=source, region_index=i, segment_index=j)
                                else:
                                    raise NotImplementedError
                        self.logger.info('\njobs ready')
                        #### IMPORTANT ####
                        # any change below may be to be also applied to `env.SlurmOverSSH.resume`
                        ###################
                        try:
                            self.env.submit_jobs()
                            self.logger.info('jobs submitted')
                            self.env.wait_for_job_completion()
                        except KeyboardInterrupt as e:
                            self.logger.critical('interrupting jobs...')
                            try:
                                ret = self.env.interrupt_jobs()
                            except KeyboardInterrupt:
                                self.logger.debug('interrupt_jobs() did not return')
                                raise e from None
                            else:
                                if not ret:
                                    raise
                        self.logger.info('jobs complete')
                        if self.env.collect_results(stage_index=s, reload_existing_rwa_files=stage.update_existing_rwa_files):
                            self.logger.info('results collected')
            finally:
                if self.env.submit_side and not self.env.debug:
                    self.env.delete_temporary_data()
                elif self.env.worker_side:
                    self.env.close()
        else:
            for stage in self._stage:
                stage(self)

    def add_collectible(self, filepath):
        """
        Registers a file as to be collected after stage completion.
        """
        self.env.collectibles.add(filepath)

    def resume(self, **kwargs):
        """
        Looks for orphaned remote jobs and collect the generated files.

        Works as a replacement for the :meth:`run` method to recover
        after connection loss.

        Recovery procedures featured by the :attr:`~tramway.analyzer.RWAnalyzer.env`
        backend may fail or recover some of the generated files only.

        See also the :meth:`~tramway.analyzer.env.environments.SlurmOverSSH.resume` method
        of the :attr:`~tramway.analyzer.RWAnalyzer.env` attribute, if available.
        """
        try:
            proc = self.env.resume
        except AttributeError:
            self.logger.error('no recovery procedure available')
        else:
            proc(**kwargs)

Attribute.register(Pipeline)


from . import stages

__all__ = ['Pipeline', 'PipelineStage', 'stages']




# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This module regroups helper functions for random walk simulations
"""

import numpy as np
import tqdm

SPACE_COLS = ['x', 'y', 'z']


def rw_is_useless(rw, nb_pos_min=2, jump_max=10):
    """Checks if the random walk is useless by checking that is has two or more
    different positions.
    """
    is_immobile = True
    is_too_big = True
    n = 0
    while 2**n <= len(rw):
        if len(rw.iloc[:2**n].x.unique()) > nb_pos_min:
            is_immobile = False
            break
        else:
            n += 1
    if np.max(np.linalg.norm(rw.loc[:, ['x', 'y']].values, axis=1)) < jump_max:
        is_too_big = False
    return is_immobile or is_too_big


def normalize_init(X, dim):
    """
    Small function to normalize the starting point of the random walk.

    Parameters
    ----------
    X : float or list, the starting point.
    dim : int, in {1,2,3}, the dimension of the random walk.
    """
    if type(X) != list:
        if X is None:
            X = 0
        return [X] * dim
    else:
        if len(X) < dim:
            return X + [0] * (dim - len(X))
        elif len(X) == dim:
            return X
        else:
            return X[:dim]


def apply_angle_dists(dist, angle, dim):
    """
    Applies a distribution of angles to a distribution of distances to generate
    the positions of the random walk in the right dimensions

    Parameters
    ----------
    dist : array of non negative values
    angle : array of angles.
    dim : the dimension of the random walk.
        If dim = 0, then angle needs to be a random array of 0 and 1.
        If dim = 1, then angle needs to be a random array of numbers between
            0 and 2pi.
        If dim = 2, then angle needs to be a 2d array, with the first line
            containing random angles between 0 and pi and the second line
            containing random angles between 0 and 2pi.
    """
    if dim == 1:
        X = dist * angle
        return X[:, np.newaxis]
    elif dim == 2:
        X = dist * np.cos(angle)
        Y = dist * np.sin(angle)
        return np.vstack((X, Y)).T
    elif dim == 3:
        X = dist * np.sin(angle[0]) * np.cos(angle[1])
        Y = dist * np.sin(angle[0]) * np.sin(angle[1])
        Z = dist * np.cos(angle[0])
        return np.vstack((X, Y, Z)).T


# # Deprecated : factor 10 gained with new implementation, which also requires
# # less space.
# def regularize_times(X_raw, t_raw, t_regular):
#     diff_t = t_regular[:, np.newaxis] - t_raw[np.newaxis, :]
#     indice_array = (np.argmin(diff_t >= 0, axis=1) - 1)
#     imin = np.argmin(indice_array)
#     if indice_array[imin] == -1:
#         t_regular = t_regular[:imin]
#         X = X_raw[indice_array[:imin]]
#     else:
#         X = X_raw[indice_array]
#     return X, t_regular


def regularize_times(X_raw, t_raw, t_regular):
    """
    This function is used to regularize times generated from some distribution
    to constant spaced times.

    Parameters
    ----------
    X_raw : numpy array, position of the random walk at different times, given
        by t_raw.
    t_raw : raw times, distribution of times : non constant delta ts.
    t_regular : the times (with constant delta t) at which we want the position
        of the random walk.

    Returns
    -------
    X : regularized X_raw, with position of the random walk at each t in
        t_regular.
    t_regular : times for which we have the position of the random walk. Can be
        shorter than the t_regular input if the maximum time in t_raw is higher
        than the maximum time in the input t_regular.
    """
    id_regu = 1
    id_raw = 0
    t_r = t_raw[id_raw]
    id_raw_into_regu = [0]
    while id_regu < len(t_regular):
        t_regu = t_regular[id_regu]
        while t_r < t_regu:
            id_raw += 1
            t_r = t_raw[id_raw]
        id_raw_into_regu.append(id_raw)
        id_regu += 1
    return X_raw[id_raw_into_regu], t_regular


# For confined movement (RW_circular_confinement)


def cart2pol(x, y):
    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)
    return rho, phi


def cart2spher(x, y, z):
    r = np.sqrt(x * x + y * y + z * z)
    phi = np.arctan2(y, x)
    theta2 = np.arccos(z / (r + 1e-8))
    return r, phi, theta2


def local_border_forces_1D(r, d_wall, sigma, V_confinement):
    f = 0
    f_slope = V_confinement/sigma
    H_f = np.heaviside(np.abs(r) - d_wall + sigma, 1)
    f = H_f * (-np.sign(r) * f_slope)
    return f


def local_border_forces_2D(r, d_wall, theta, sigma, V_confinement):
    f = np.zeros((2,))
    f_slope = V_confinement / sigma
    H_f = np.heaviside(r - d_wall + sigma, 1)
    f[0] = H_f * (-r * np.cos(theta) * f_slope)
    f[1] = H_f * (-r * np.sin(theta) * f_slope)
    return f


def local_border_forces_3D(r, d_wall, theta, phi, sigma, V_confinement):
    f = np.zeros((3,))
    f_slope = V_confinement/sigma
    H_f = np.heaviside(r - d_wall + sigma, 1)
    f[0] = H_f * (- r * np.sin(theta) * np.cos(phi) * f_slope)
    f[1] = H_f * (- r * np.sin(theta) * np.sin(phi) * f_slope)
    f[2] = H_f * (- r * np.cos(theta) * f_slope)
    return f

# For Self Avoiding Random Walks

# Basic Functions


def pluss(pt_1, pt_2, dim):
    pt = []
    for coordinate in range(dim):
        pt.append(pt_1[coordinate] + pt_2[coordinate])
    return tuple(pt)


def moinss(pt_1, pt_2, dim):
    pt = []
    for coordinate in range(dim):
        pt.append(pt_1[coordinate] - pt_2[coordinate])
    return tuple(pt)


def norme_manhattan(pt):
    resul = 0
    for coord in pt:
        resul += abs(coord)
    return resul


def enleve_coord(pt, coordinate):
    pt = list(pt)
    pt.pop(coordinate)
    return tuple(pt)


def voisin(pt, coord, ecart):
    resul = list(pt)
    resul[coord] = resul[coord] + ecart
    return tuple(resul)


def intersection(ensemble_1, ensemble_2):
    resul = []
    for v in ensemble_1:
        if v in ensemble_2:
            resul.append(v)
    return resul


def distance_euclidean(pt1, pt2, dim):
    resul = 0
    for coordinate in range(dim):
        resul += ((pt1[coordinate] - pt2[coordinate]) ** 2)
    return np.sqrt(resul)


# Create network


def create_minmax(dim, Petite_Origine):
    les_minmax = []
    for i in range(dim):
        les_minmax.append({Petite_Origine: [0, 0]})
    return les_minmax


def create_deplacement(dim):
    deplacements = []
    for i in range(dim):
        deplacements.append(tuple(([0] * i) + [1] + ([0] * (dim - (i + 1)))))
        deplacements.append(tuple(([0] * i) + [-1] + ([0] * (dim - (i + 1)))))
    return deplacements


def construct_chemin(graphe, distance, v):
    ch = [v]
    d = distance[v]
    x = v
    while d > 0:
        for y in graphe[x]:
            if distance[y] == d - 1:
                ch.append(y)
                d -= 1
                x = y
                break
    return ch


def chemin(depart, arrivees, graphe):
    distance = {}
    for vertice in graphe:
        distance[vertice] = None
    distance[depart] = 0
    frontiere = [depart]
    k = 1
    while frontiere != []:
        new_front = []
        for vertice in frontiere:
            for vertice_2 in graphe[vertice]:
                if vertice_2 in arrivees:
                    distance[vertice_2] = k
                    return construct_chemin(graphe, distance, vertice_2)
                if distance[vertice_2] is None:
                    distance[vertice_2] = k
                    new_front.append(vertice_2)
        k += 1
        frontiere = new_front


def calcul_proba(tab, bias, dim):
    resul = []
    for coordinate in range(len(tab)):
        if tab[coordinate] >= 0:
            if bias == 0:
                resul.append(coordinate)
            elif bias > 0:
                resul.extend([coordinate] * (bias - 1 + tab[coordinate]))
            else:
                resul.extend([coordinate] * (-bias + (2 * dim) - 1 -
                                             tab[coordinate]))
    return resul


# For Random Walks on Pattern : Pattern creation


PATTERN_SIERPINSKI = {
    "classic": np.array([[1, 1, 0],
                         [1, 1, 1],
                         [0, 1, 0]]),
    "butterfly": np.array([[1, 1, 0],
                           [1, 1, 1],
                           [0, 1, 1]]),
    "stairs": np.array([[1, 0, 0],
                        [1, 1, 0],
                        [1, 1, 1]]),
    "cross5": np.array([[1, 1, 1, 1, 1],
                        [1, 0, 1, 0, 1],
                        [1, 1, 1, 1, 1],
                        [1, 0, 1, 0, 1],
                        [1, 1, 1, 1, 1]]),
    "square5": np.array([[0, 1, 1, 1, 0],
                         [0, 1, 1, 1, 0],
                         [1, 1, 1, 1, 1],
                         [0, 1, 1, 1, 0],
                         [0, 1, 1, 1, 0]]),
    "square5_full": np.array([[0, 1, 1, 1, 0],
                              [1, 0, 1, 0, 1],
                              [1, 1, 1, 1, 1],
                              [1, 0, 1, 0, 1],
                              [0, 1, 1, 1, 0]]),
    "cheese5": np.array([[1, 0, 1, 0, 1],
                         [1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 1],
                         [1, 0, 1, 0, 1]])
}


def sierpinski_carpet_deterministic_generator(global_iteration=4,
                                              nature="classic",
                                              path_output=None):
    base_pattern = PATTERN_SIERPINSKI[nature]
    base_size = base_pattern.shape[0]
    n_2 = int(np.floor(base_size/2))
    matrice_zeros = np.zeros((base_size, base_size), dtype=int)
    pattern = base_pattern
    for _ in range(global_iteration):
        ll = pattern.shape[0]
        pattern_new = np.zeros((base_size * ll, base_size * ll))
        for i in range(ll):
            for j in range(ll):
                i_mean = (i)*base_size + base_size - (n_2+1)
                j_mean = (j)*base_size + base_size - (n_2+1)
                if pattern[i, j] == 0:
                    pattern_new[i_mean-n_2:i_mean+n_2+1,
                                j_mean-n_2:j_mean+n_2+1] = matrice_zeros
                else:
                    pattern_new[i_mean-n_2:i_mean+n_2+1,
                                j_mean-n_2:j_mean+n_2+1] = base_pattern
        pattern = pattern_new
    if path_output is not None:
        np.save(path_output, pattern)
    return pattern


def generate_the_fractals_and_save_them(DIR, types=[(7, "classic"),
                                                    (7, "butterfly"),
                                                    (5, "cross5"),
                                                    (5, "square5_full"),
                                                    (5, "cheese5")]):
    for (num_iter, type_name) in types:
        sierpinski_carpet_deterministic_generator(
            global_iteration=num_iter,
            nature=type_name,
            path_output=f'{DIR}\{type_name}')


def update_position(motif, i_init, j_init):
    keep_going = True
    counter = 0
    output_loc = np.zeros((4, 2), dtype=int)

    if motif[i_init+1, j_init] == 1:
        output_loc[counter, 0] = i_init+1
        output_loc[counter, 1] = j_init
        counter += 1
    if motif[i_init, j_init+1] == 1:
        output_loc[counter, 0] = i_init
        output_loc[counter, 1] = j_init+1
        counter += 1
    if motif[i_init-1, j_init] == 1:
        output_loc[counter, 0] = i_init-1
        output_loc[counter, 1] = j_init
        counter += 1
    if motif[i_init, j_init-1] == 1:
        output_loc[counter, 0] = i_init
        output_loc[counter, 1] = j_init-1
        counter += 1

    if counter == 0:
        k = 0
        l = 0
        keep_going = False
    else:
        counter_alea = np.random.randint(counter)
        k = output_loc[counter_alea, 0]
        l = output_loc[counter_alea, 1]
    return k, l, keep_going


# For random walks on trees. (Diffusion-limited aggregation)

def update_step(i, j):
    p_i_or_j = np.random.rand()
    move = np.random.randint(0, 2) * 2 - 1
    if p_i_or_j < 0.5:
        i += move
    else:
        j += move
    return i, j


def define_seed(tree):
    x, y = zip(*tree.keys())
    x_min, x_max, y_min, y_max = min(x), max(x), min(y), max(y)
    xc, yc = (x_min+x_max)*1/2, (y_min+y_max)*1/2
    r = np.power(np.power((x_max-x_min)/2, 2) +
                 np.power((y_max-y_min)/2, 2), 0.5)
    return (xc, yc), r


def get_initial_pos_on_circle(xc, yc, r):
    theta = 2. * np.pi * np.random.random()
    x0 = int(xc + r*np.cos(theta))
    y0 = int(yc + r*np.sin(theta))
    return x0, y0


def rw(tree, epsilon=5):
    (xc, yc), r = define_seed(tree)
    x, y = get_initial_pos_on_circle(xc, yc, r)
    while True:
        x, y = update_step(x, y)
        if np.sqrt((x-xc)**2 + (y-yc)**2) > r + epsilon:
            break
        if (((x-1, y) in tree) or ((x+1, y) in tree) or
                ((x, y-1) in tree) or ((x, y+1) in tree)):
            tree[(x, y)] = len(tree)
            break
    return tree


def get_raw_size(tree):
    list_tree = np.array(list(tree), dtype=int)
    i_min, i_max = np.min(list_tree[:, 0]), np.max(list_tree[:, 0])
    j_min, j_max = np.min(list_tree[:, 1]), np.max(list_tree[:, 1])
    return int(np.floor(np.max([j_max - j_min + 1, i_max - i_min + 1])))


def grow(tree=None, n_trial_max=100, n_eff_min=100):
    if tree is None:
        tree = {(0, 0): 0}
    i = 0
    pbar = tqdm.tqdm_notebook()
    while (i < n_trial_max and get_raw_size(tree) < n_eff_min):
        tree = rw(tree)
        i += 1
        pbar.update(1)
    pbar.close()
    print(f'n_eff = {get_raw_size(tree)}, i = {i}')
    return tree


def generate_the_DLA(n_trial_max=1000, growth=1.25, n_eff_min=100):
    tree = grow(tree=None, n_trial_max=n_trial_max, n_eff_min=n_eff_min)
    list_tree = np.array(list(tree), dtype=int)
    n_eff = int(get_raw_size(tree) * growth)
    n_eff_2 = int(np.floor(n_eff/2))
    DLA_array = np.zeros((n_eff, n_eff))
    list_tree = list_tree + [n_eff_2, n_eff_2]
    DLA_array[list_tree[:, 0], list_tree[:, 1]] = 1
    return DLA_array


def generate_and_save_DLA(DIR, name_output="DLA", n_trial_max=1000,
                          n_eff_min=100, growth=1.25):
    DLA_array = generate_the_DLA(n_trial_max=n_trial_max, growth=growth,
                                 n_eff_min=n_eff_min)
    np.save(f'{DIR}\\{name_output}', DLA_array)



# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This module is where functions creating multiple random walks at a time are
defined.
"""

import multiprocessing as mp

import numpy as np
import pandas as pd
import tqdm

from .rw_simulation import *
from .rw_misc import rw_is_useless


def generate_random_number(type_n, type_gen, a, b=None, p=None):
    """
    Useful function that outputs a randomly generated number or string
    depending on the given parameters

    Parameters
    ----------
    type_n : string, the type of the random number generated.
    type_gen : string, the type of random distribution chosen.
    a : number, parameter relevant to the chosen distribution.
    b : number, optional depending on the type of generation chosen.
    """
    if type_n == 'float':
        if type_gen == 'uni':
            return np.random.rand() * (b-a) + a
        elif type_gen == 'exp':
            return 10**(np.random.rand() * (np.log10(b)-np.log10(a)) +
                        np.log10(a))
        else:
            raise TypeError(f'Unrecognized type : {type_gen}')
    elif type_n == 'int':
        if type_gen == 'uni':
            return np.random.randint(a, b+1)
        elif type_gen == 'exp':
            return (10**(np.random.rand() * (np.log10(b)-np.log10(a)) +
                         np.log10(a))).astype(int)
        else:
            raise TypeError(f'Unrecognized type : {type_gen}')
    elif type_n == 'str':
        i = np.random.choice(len(a), p=p)
        return a[i]
    elif type_n == 'cst':
        return a
    else:
        raise TypeError(f'Unrecognized number type : {type_n}')


def create_random_rw(args):
    """Returns a tuple of the simulated random walk and the parameters of the
    simulation a probability distribution of possible random walks.

    Parameters
    ----------
    args : tuple of (ps (list or None or int, should be accepted by
        np.random.choice as the p parameter), types (see
        "rw_feature_generator")).

    Returns
    -------
    rw : pandas DataFrame of the position/time of the simulated random walk.
    rw_dict_prms : dict of the parameters of the simulated random walk.
    """
    ps, types, nb_pos_min, jump_max, id_traj = args
    rw_id_type = np.random.choice(len(types), p=ps)
    rw_dict_prms = {}
    for prm, val in types[rw_id_type][1].items():
        rw_dict_prms[prm] = generate_random_number(*val)
    rw = types[rw_id_type][0](**rw_dict_prms)
    while rw_is_useless(rw, nb_pos_min, jump_max):
        rw_dict_prms = {}
        for prm, val in types[rw_id_type][1].items():
            rw_dict_prms[prm] = generate_random_number(*val)
        rw = types[rw_id_type][0](**rw_dict_prms)
    rw['n'] = id_traj
    rw_dict_prms['func_name'] = types[rw_id_type][0].__name__
    return (rw, rw_dict_prms)


def rw_feature_generator(n, types=[(RW_gauss_dist,
                                    {'d_l': ('float', 'exp', 0.01, 0.1),
                                     'T_max': ('float', 'exp', 0.1, 1)})],
                         ps=[1], nb_pos_min=3, jump_max=10, **kwargs):
    """
    Generator of random walks.

    Parameters
    ----------
    n : int, the number of random walks we want to generate.
    type : list that describes which random walks we generate.
        Each item is a tuple with 2 elements.
        - The first is the function (see rw_simulation) which characterizes
        which type of random walk it produces.
        - The second is a dictionary whose keys are the parameters we want to
        pass to the random walk function and values the parameters describing
        how those parameter values are generated with `generate_random_number`.
    ps : parameter passed to `np.random.choice` as p, should have same
        dimension as type. Controls the probability distribution of choosing
        some type.

    Yields
    ------
    A tuple of :
        - the list of functions used to extract features out of the random walk
        - the id of the random walk generated
        - rw, the random walk pandas.DataFrame object
        - ns, the feature extraction parameter
        - rw_dict_prms, the dictionary of parameters relevant to the random
        walk generation.
        - get_rw, the bool describing if we want to eventually return the
        random walk.
    """
    for i in range(n):
        yield (create_random_rw((ps, types, nb_pos_min, jump_max, i)) +
               (kwargs,))


def create_batch_rw(n=100, ps=[1], nb_process=4, nb_pos_min=2, jump_max=10,
                    chunksize=10, pbar=True,
                    types=[(RW_gauss_dist,
                            {'d_l': ('float', 'exp', 0.01, 0.1),
                             'T_max': ('float', 'exp', 0.1, 1)})]):
    """Function which uses multiprocessing to rapidly create multiple random
    walks.

    Parameters
    ----------
    n : int, the number of random walks we want to generate.
    ps : parameter passed to `np.random.choice` as p, should have same
        dimension as type. Controls the probability distribution of choosing
        some type.
    nb_process : number of CPU processes used simultaneous (makes use of
        multiprocessing built-in Python module).
    type : list that describes which random walks we generate.
        Each item is a tuple with 2 elements.
        - The first is the function (see rw_simulation) which characterizes
        which type of random walk it produces.
        - The second is a dictionary whose keys are the parameters we want to
        pass to the random walk function and values the parameters describing
        how those parameter values are generated with `generate_random_number`.

    Returns
    -------
    - pandas DataFrame of the simulated random walks.
    - dictionary of dictionaries. Each key is the index of the random walk,
        each value is a dictionary of the values taken by the parameters of the
        random walk.
    """
    prms = [(ps, types, nb_pos_min, jump_max, i) for i in range(n)]
    if pbar:
        if nb_process is None:
            raw_data = list(map(create_random_rw,
                                tqdm.tqdm_notebook(prms,
                                                   desc='generating RWs')))
        else:
            with mp.Pool(nb_process) as p:
                raw_data = list(tqdm.tqdm_notebook(
                    p.imap(create_random_rw, prms, chunksize),
                    total=n, desc='generating RWs'))
    else:
        if nb_process is None:
            raw_data = list(map(create_random_rw, prms))
        else:
            with mp.Pool(nb_process) as p:
                raw_data = list(p.imap(create_random_rw, prms, chunksize))
    RW_data = []
    RW_prm = {}
    for traj, rw in enumerate(raw_data):
        RW_prm[traj] = rw[1]
        RW_data.append(rw[0])
    return pd.concat(RW_data).reset_index(drop=True), RW_prm



# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute.abc import Attribute, abstractmethod

class Mapper(Attribute):
    """
    Abstract base class for the :attr:`~tramway.analyzer.RWAnalyzer.mapper` attribute
    of an :class:`~tramway.analyzer.RWAnalyzer` object.
    """
    @abstractmethod
    def infer(self, sampling):
        pass




# -*- coding: utf-8 -*-

# Copyright © 2017-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import ChainArray
from .base import *
from .gradient import *
from .standard_d import interruptible_minimize
from warnings import warn
from math import pi, log
import numpy as np
import pandas as pd
#from scipy.optimize import minimize
from collections import OrderedDict
import time


setup = {'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see also sigma; default is 0.03)'))),
        ('diffusivity_prior',   ('-d', dict(type=float, help='prior on the diffusivity'))),
        ('potential_prior',     ('-v', dict(type=float, help='prior on the potential'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')),
        ('max_iter',        dict(type=int, help='maximum number of iterations (~100)')),
        ('compatibility',       ('-c', '--inferencemap', '--compatible',
                    dict(action='store_true', help='InferenceMAP compatible'))),
        ('rgrad',       dict(help="alternative gradient for the regularization; can be 'delta0' or 'delta1'")),
        ('export_centers',      dict(action='store_true')),
        ('verbose',         ()))),
    'cell_sampling': 'connected'}
setup_with_grad_arguments(setup)


class DV(ChainArray):
    """
    Convenience class for handling the D+V parameter vector and related priors.

    Objects of this class are not supposed to be serialized in .rwa files.
    """
    __slots__ = ('_diffusivity_prior', '_potential_prior', 'minimum_diffusivity', 'prior_include')

    def __init__(self, diffusivity, potential, diffusivity_prior=None, potential_prior=None, \
        minimum_diffusivity=None, positive_diffusivity=None, prior_include=None):
        # positive_diffusivity is for backward compatibility
        ChainArray.__init__(self, 'D', diffusivity, 'V', potential)
        self._diffusivity_prior = diffusivity_prior
        self._potential_prior = potential_prior
        self.minimum_diffusivity = minimum_diffusivity
        if minimum_diffusivity is None and positive_diffusivity is True:
            self.minimum_diffusivity = 0
        self.prior_include = prior_include

    @property
    def D(self):
        return self['D']

    @property
    def V(self):
        return self['V']

    @D.setter
    def D(self, diffusivity):
        self['D'] = diffusivity

    @V.setter
    def V(self, potential):
        self['V'] = potential

    def diffusivity_prior(self, j):
        if self._diffusivity_prior and (self.prior_include is None or self.prior_include[j]):
            return self._diffusivity_prior
        else:
            return None

    def potential_prior(self, j):
        if self._potential_prior and (self.prior_include is None or self.prior_include[j]):
            return self._potential_prior
        else:
            return None



def dv_neg_posterior(x, dv, cells, sigma2, jeffreys_prior, dt_mean, \
        index, reverse_index, grad_kwargs, y0, verbose, posteriors):
    """
    Adapted from InferenceMAP's *dvPosterior* procedure modified:

    .. code-block:: c++

        for (int a = 0; a < NUMBER_OF_ZONES; a++) {
            ZONES[a].gradVx = dvGradVx(DV,a);
            ZONES[a].gradVy = dvGradVy(DV,a);
            ZONES[a].gradDx = dvGradDx(DV,a);
            ZONES[a].gradDy = dvGradDy(DV,a);
            ZONES[a].priorActive = true;
        }


        for (int z = 0; z < NUMBER_OF_ZONES; z++) {
            const double gradVx = ZONES[z].gradVx;
            const double gradVy = ZONES[z].gradVy;
            const double gradDx = ZONES[z].gradDx;
            const double gradDy = ZONES[z].gradDy;

            const double D = DV[2*z];

            for (int j = 0; j < ZONES[z].translocations; j++) {
                const double dt = ZONES[z].dt[j];
                const double dx = ZONES[z].dx[j];
                const double dy = ZONES[z].dy[j];
                const double  Dnoise = LOCALIZATION_ERROR*LOCALIZATION_ERROR/dt;

                result += - log(4.0*PI*(D + Dnoise)*dt) - ((dx-D*gradVx*dt)*(dx-D*gradVx*dt) + (dy-D*gradVy*dt)*(dy-D*gradVy*dt))/(4.0*(D+Dnoise)*dt);
            }

            if (ZONES[z].priorActive == true) {
                result -= V_PRIOR*(gradVx*gradVx*ZONES[z].areaX + gradVy*gradVy*ZONES[z].areaY);
                result -= D_PRIOR*(gradDx*gradDx*ZONES[z].areaX + gradDy*gradDy*ZONES[z].areaY);
                if (JEFFREYS_PRIOR == 1) {
                    result += 2.0*log(D*1.00) - 2.0*log(D*ZONES[z].dtMean + LOCALIZATION_ERROR*LOCALIZATION_ERROR);
            }
        }

    with ``dx-D*gradVx*dt`` and ``dy-D*gradVy*dt`` modified as ``dx+D*gradVx*dt`` and ``dy+D*gradVy*dt`` respectively.

    See also :ref:`DV <inference_dv>`.
    """
    if verbose:
        t = time.time()

    # extract `D` and `V`
    dv.update(x)
    D = dv.D
    V = dv.V
    #

    if dv.minimum_diffusivity is not None:
        observed_min = np.min(D)
        if observed_min < dv.minimum_diffusivity and not \
                np.isclose(observed_min, dv.minimum_diffusivity):
            warn(DiffusivityWarning(observed_min, dv.minimum_diffusivity))
    noise_dt = sigma2

    # for all cell
    raw_posterior = priors = 0.
    for j, i in enumerate(index):
        cell = cells[i]
        n = len(cell) # number of translocations

        # spatial gradient of the local potential energy
        gradV = cells.grad(i, V, reverse_index, **grad_kwargs)
        #print('{}\t{}\t{}\t{}\t{}\t{}'.format(i+1,D[j], V[j], -gradV[0], -gradV[1], result))
        #print('{}\t{}\t{}'.format(i+1, *gradV))
        if gradV is None:
            continue

        # various posterior terms
        #print(cell.dt)
        D_dt = D[j] * cell.dt
        denominator = 4. * (D_dt + noise_dt)
        dr_minus_drift = cell.dr + np.outer(D_dt, gradV)
        # non-directional squared displacement
        ndsd = np.sum(dr_minus_drift * dr_minus_drift, axis=1)
        res = n * log(pi) + np.sum(np.log(denominator)) + np.sum(ndsd / denominator)
        if np.isnan(res):
            #print('isnan')
            continue
        raw_posterior += res

        # priors
        potential_prior = dv.potential_prior(j)
        if potential_prior:
            priors += potential_prior * cells.grad_sum(i, gradV * gradV, reverse_index)
        diffusivity_prior = dv.diffusivity_prior(j)
        if diffusivity_prior:
            # spatial gradient of the local diffusivity
            gradD = cells.grad(i, D, reverse_index, **grad_kwargs)
            if gradD is not None:
                # `grad_sum` memoizes and can be called several times at no extra cost
                priors += diffusivity_prior * cells.grad_sum(i, gradD * gradD, reverse_index)
        #print('{}\t{}\t{}'.format(i+1, D[j], result))
    if jeffreys_prior:
        priors += 2. * np.sum(np.log(D * dt_mean + sigma2) - np.log(D))

    result = raw_posterior + priors
    posteriors.append([raw_posterior, result])

    if verbose:
        print('objective: {}\t time: {}ms'.format(result, int(round((time.time() - t) * 1e3))))

    return result - y0


def dv_neg_posterior1(x, dv, cells, sigma2, jeffreys_prior, dt_mean, \
        index, reverse_index, grad_kwargs, y0, verbose, posteriors):
    """
    Similar to :func:`dv_neg_posterior`.
    The smoothing priors feature an alternative spatial "gradient" implemented using
    :meth:`~tramway.inference.base.Distributed.local_variation` instead of
    :meth:`~tramway.inference.base.Distributed.grad`.
    """
    if verbose:
        t = time.time()

    # extract `D` and `V`
    dv.update(x)
    D = dv.D
    V = dv.V
    #

    if dv.minimum_diffusivity is not None:
        observed_min = np.min(D)
        if observed_min < dv.minimum_diffusivity and \
                not np.isclose(observed_min, dv.minimum_diffusivity):
            warn(DiffusivityWarning(observed_min, dv.minimum_diffusivity))
    noise_dt = sigma2

    # for all cell
    raw_posterior = priors = 0.
    for j, i in enumerate(index):
        cell = cells[i]
        n = len(cell) # number of translocations

        # spatial gradient of the local potential energy
        gradV = cells.grad(i, V, reverse_index, **grad_kwargs)
        #print('{}\t{}\t{}\t{}\t{}\t{}'.format(i+1,D[j], V[j], -gradV[0], -gradV[1], result))
        #print('{}\t{}\t{}'.format(i+1, *gradV))
        if gradV is None:
            continue

        # various posterior terms
        #print(cell.dt)
        D_dt = D[j] * cell.dt
        denominator = 4. * (D_dt + noise_dt)
        dr_minus_drift = cell.dr + np.outer(D_dt, gradV)
        # non-directional squared displacement
        ndsd = np.sum(dr_minus_drift * dr_minus_drift, axis=1)
        res = n * log(pi) + np.sum(np.log(denominator)) + np.sum(ndsd / denominator)
        if np.isnan(res):
            #print('isnan')
            continue
        raw_posterior += res

        # priors
        potential_prior = dv.potential_prior(j)
        if potential_prior:
            # spatial variation of the local potential
            deltaV = cells.local_variation(i, V, reverse_index, **grad_kwargs)
            if deltaV is not None:
                priors += potential_prior * cells.grad_sum(i, deltaV * deltaV, reverse_index)
        diffusivity_prior = dv.diffusivity_prior(j)
        if diffusivity_prior:
            # spatial variation of the local diffusivity
            deltaD = cells.local_variation(i, D, reverse_index, **grad_kwargs)
            if deltaD is not None:
                # `grad_sum` memoizes and can be called several times at no extra cost
                priors += diffusivity_prior * cells.grad_sum(i, deltaD * deltaD, reverse_index)
        #print('{}\t{}\t{}'.format(i+1, D[j], result))
    if jeffreys_prior:
        priors += 2. * np.sum(np.log(D * dt_mean + sigma2) - np.log(D))

    result = raw_posterior + priors
    if posteriors is not None:
        posteriors.append([raw_posterior, result])

    if verbose:
        print('objective: {}\t time: {}ms'.format(result, int(round((time.time() - t) * 1e3))))

    return result - y0



def inferDV(cells, diffusivity_prior=None, potential_prior=None, \
    jeffreys_prior=False, min_diffusivity=None, max_iter=None, epsilon=None, \
    export_centers=False, verbose=True, compatibility=False, \
    D0=None, V0=None, rgrad=None, **kwargs):

    localization_error = cells.get_localization_error(kwargs, 0.03, True)

    # initial values
    index, reverse_index, n, dt_mean, D_initial, min_diffusivity, D_bounds, border = \
        smooth_infer_init(cells, min_diffusivity=min_diffusivity, jeffreys_prior=jeffreys_prior,
        sigma2=localization_error)
    # V initial values
    if V0 is None:
        try:
            if compatibility:
                raise Exception # skip to the except block
            volume = [ cells[i].volume for i in index ]
        except:
            V_initial = -np.log(n / np.max(n))
        else:
            density = n / np.array([ np.inf if v is None else v for v in volume ])
            density[density == 0] = np.min(density[0 < density])
            V_initial = np.log(np.max(density)) - np.log(density)
    else:
        if np.isscalar(V0):
            V_initial = np.full(D_initial.size, V0)
        elif V0.size == D_initial.size:
            V_initial = V0
        else:
            raise ValueError('wrong size for V0')
    if D0 is not None:
        if np.isscalar(D0):
            D_initial[...] = D0
        elif D0.size == D_initial.size:
            D_initial = D0
        else:
            raise ValueError('wrong size for D0')

    dv = DV(D_initial, V_initial, diffusivity_prior, potential_prior, min_diffusivity)
    posteriors = []

    # gradient options
    grad_kwargs = get_grad_kwargs(kwargs, epsilon=epsilon, compatibility=compatibility)

    # parametrize the optimization algorithm
    default_lBFGSb_options = dict(maxiter=1e3, maxfun=1e10, maxcor=dv.combined.size, ftol=1e-8)
    # in L-BFGS-B the number of iterations is usually very low (~10-100) while the number of
    # function evaluations is much higher (~1e4-1e5);
    # with maxfun defined, an iteration can stop anytime and the optimization may terminate
    # with an error message
    if min_diffusivity is None:
        bounds = None
        options = {}
    else:
        V_bounds = [(None, None)] * V_initial.size
        bounds = D_bounds + V_bounds
        options = dict(default_lBFGSb_options)
    _kwargs = {}
    for k in ('jac',):
        try:
            _kwargs[k] = kwargs.pop(k)
        except KeyError:
            pass
    options.update(kwargs.pop('options', {}))
    options.update(**kwargs) # for backward compatibility
    if max_iter:
        options['maxiter'] = max_iter
    if verbose:
        options['disp'] = verbose
    if options:
        _kwargs['options'] = options

    # posterior function
    if rgrad in ('delta','delta0','delta1'):
        fun = dv_neg_posterior1
    else:
        if rgrad not in (None, 'grad', 'grad1', 'gradn'):
            warn('unsupported rgrad: {}'.format(rgrad), RuntimeWarning)
        fun = dv_neg_posterior

    # posterior function input arguments
    args = (dv, cells, localization_error, jeffreys_prior, dt_mean,
            index, reverse_index, grad_kwargs)

    # get the initial posterior value so that it is subtracted from the further evaluations
    y0 = fun(dv.combined, *(args + (0., False, [])))
    if verbose:
        print('At X0\tactual posterior= {}\n'.format(y0))
    #y0 = 0.
    args = args + (y0, 1 < int(verbose), posteriors)

    # run the optimization routine
    result = interruptible_minimize(fun, dv.combined, args=args, bounds=bounds, **_kwargs)
    if not (result.success or verbose):
        warn('{}'.format(result.message), OptimizationWarning)

    y = np.array(result.x)

    # collect the result
    dv.update(y)
    D, V = dv.D, dv.V
    if np.any(V < 0):
        V -= np.min(V)
    DVF = pd.DataFrame(np.stack((D, V), axis=1), index=index, \
        columns=[ 'diffusivity', 'potential'])

    # derivate the forces
    index_, F = [], []
    for i in index:
        gradV = cells.grad(i, V, reverse_index, **grad_kwargs)
        if gradV is not None:
            index_.append(i)
            F.append(-gradV)
    if F:
        F = pd.DataFrame(np.stack(F, axis=0), index=index_, \
            columns=[ 'force ' + col for col in cells.space_cols ])
    else:
        warn('not any cell is suitable for evaluating the local force', RuntimeWarning)
        F = pd.DataFrame(np.zeros((0, len(cells.space_cols)), dtype=V.dtype), \
            columns=[ 'force ' + col for col in cells.space_cols ])
    DVF = DVF.join(F)

    # add extra information if required
    if export_centers:
        xy = np.vstack([ cells[i].center for i in index ])
        DVF = DVF.join(pd.DataFrame(xy, index=index, \
            columns=cells.space_cols))
        #DVF.to_csv('results.csv', sep='\t')

    # format the posteriors
    posteriors = pd.DataFrame(np.array(posteriors), columns=['fit', 'total'])

    return DVF, posteriors

